"""Clustering from sphere parents."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import networkx as nx
import pandas as pd
from scipy.spatial import KDTree
from scipy.spatial.distance import pdist

from axon_synthesis.inputs.clustering.utils import common_path
from axon_synthesis.utils import COORDS_COLS


def _check_cluster(a, b, nodes, terminal_nodes, pair_paths, cluster_ids, max_path_distance) -> None:
    """Check that 2 points are in the same cluster."""
    term_a = terminal_nodes.iloc[a].name
    term_b = terminal_nodes.iloc[b].name
    if term_a == -1 or term_b == -1:
        return
    try:
        path = pair_paths[term_a][term_b]
    except KeyError:
        path = pair_paths[term_b][term_a]

    if pdist(nodes.loc[path, COORDS_COLS].to_numpy()).max() > max_path_distance:
        # Skip if a point on the path exceeds the clustering distance
        return

    # TODO: Do not cluster the terminals if they are in different regions?
    # Or if the path between them goes too far inside another region?

    # Add points to clusters
    term_a_cluster_id = cluster_ids.loc[term_a]
    term_b_cluster_id = cluster_ids.loc[term_b]

    # If term_a is already in a cluster
    if term_a_cluster_id != -1:
        # If term_b is also already in a cluster
        if term_b_cluster_id != -1:
            # Transfer all terminals from the cluster of term_b to the one of term_a
            cluster_ids.loc[cluster_ids == term_b_cluster_id] = term_a_cluster_id
        else:
            # Add term_b to the cluster of term_a
            cluster_ids.loc[term_b] = term_a_cluster_id
    else:  # noqa: PLR5501
        # If term_b is already in a cluster
        if term_b_cluster_id != -1:
            # Add term_a to the cluster of term_b
            cluster_ids.loc[term_a] = term_b_cluster_id
        else:
            # Create new cluster
            cluster_ids.loc[[term_a, term_b]] = cluster_ids.max() + 1


def compute_clusters(  # noqa: PLR0913
    config,
    config_name,
    axon_id,
    group_name,
    group_path,
    group,
    nodes,
    edges,
    directed_graph,
    shortest_paths,
    node_to_terminals,
    **_kwargs,
):
    """All parents up to the common ancestor must be inside the sphere to be merged."""
    sphere_radius = config["sphere_radius"]
    max_path_distance = config.get("max_path_distance", sphere_radius)

    # Get the complete morphology
    new_terminal_points: list[list] = []

    graph: nx.Graph = nx.Graph(directed_graph)
    terminals = nodes.loc[nodes["is_terminal"]]

    pair_paths = dict(i for i in nx.all_pairs_shortest_path(graph) if i[0] in set(terminals.index))

    # Get the pairs of terminals closer to the given distance
    terminal_tree = KDTree(terminals[COORDS_COLS].to_numpy())
    terminal_pairs = terminal_tree.query_pairs(sphere_radius)

    # Initialize cluster IDs
    cluster_ids = pd.Series(-1, index=nodes.index.to_numpy())

    # The root can not be part of a cluster so it is given a specific cluster ID
    cluster_ids.loc[-1] = cluster_ids.max() + 1  # type: ignore[call-overload]
    group.loc[group["terminal_id"] == 0, "tuft_id"] = cluster_ids.loc[-1]

    # Check that the paths between each pair do not exceed the given distance
    for a, b in terminal_pairs:
        _check_cluster(a, b, nodes, terminals, pair_paths, cluster_ids, max_path_distance)

    # Create cluster IDs for not clustered terminals
    not_clustered_mask = (nodes["is_terminal"]) & (cluster_ids == -1)
    cluster_ids.loc[not_clustered_mask] = (
        cluster_ids.loc[not_clustered_mask].reset_index(drop=True).index + cluster_ids.max() + 1
    )

    # Reset cluster IDs to consecutive values
    nodes["tuft_id"] = cluster_ids.map(
        {v: k for k, v in enumerate(sorted(cluster_ids.unique()), start=-1)},
    )

    # Groupby points by cluster IDs
    clusters = nodes.loc[nodes["is_terminal"]].groupby("tuft_id")
    sorted_clusters = sorted(clusters, key=lambda x: x[1].size, reverse=True)

    new_terminal_id = 1

    # Ensure there are at least 4 clusters
    for num_cluster, (_, cluster) in enumerate(sorted_clusters):
        # Compute the common ancestor
        cluster_common_path = common_path(
            directed_graph,
            cluster.index.to_list(),
            shortest_paths=shortest_paths,
        )
        common_ancestors = [cluster_common_path[-1]]

        # Compute the number of missing clusters
        missing_points = max(
            0,
            3 - (len(sorted_clusters[num_cluster + 1 :]) + len(new_terminal_points)),
        )
        is_root = -1 in cluster.index
        if is_root:
            # The root node can not be split
            missing_points = 0

        # Split the cluster if needed in order to get at least 3 points
        common_ancestor_ind = 0
        while 0 < len(common_ancestors) <= missing_points:
            if common_ancestor_ind >= len(common_ancestors):
                break
            sub_ancestors = edges.loc[
                edges["source"] == common_ancestors[common_ancestor_ind],
                "target",
            ].tolist()
            if sub_ancestors:
                common_ancestors.pop(common_ancestor_ind)
                common_ancestors.extend(sub_ancestors)
            else:
                common_ancestor_ind += 1

        # Add the points of the cluster
        for common_ancestor in common_ancestors:
            terminals_with_current_ancestor = cluster.loc[
                list(set(node_to_terminals[common_ancestor]).intersection(cluster.index))
            ].sort_index()
            new_terminal_points.append(
                [
                    group_name,
                    group_path,
                    config_name,
                    axon_id,
                    new_terminal_id if not is_root else 0,
                    len(terminals_with_current_ancestor),
                    *terminals_with_current_ancestor[COORDS_COLS].mean().tolist(),
                ],
            )
            if not is_root:
                group.loc[
                    group["graph_node_id"].isin(terminals_with_current_ancestor.index),
                    "tuft_id",
                ] = new_terminal_id
                new_terminal_id += 1

    return new_terminal_points, group["tuft_id"]



"""Tools to create the main trunk of an axon."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#



"""Configuration file for the Sphinx documentation builder."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

from importlib import metadata

# -- Project information -----------------------------------------------------

project_name = "Axon Synthesis"
package_name = "axon-synthesis"

# The short X.Y version
version = metadata.version(package_name)

# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_click",
]

todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx-bluebrain-theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_theme_options = {
    "metadata_distribution": package_name,
}

html_title = project_name

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# autosummary settings
autosummary_generate = True

# autodoc settings
autodoc_typehints = "signature"
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
}

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
    "neurots": ("https://neurots.readthedocs.io/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "python": ("https://docs.python.org/3", None),
}

# MyST parser settings
myst_enable_extensions = []
myst_heading_anchors = 5
myst_all_links_external = True
suppress_warnings = ["myst.header"]



"""Tests for the axon_synthesis.cli module."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from click import Context

import axon_synthesis.cli


def test_cli(cli_runner):
    """Test the CLI."""
    result = cli_runner.invoke(axon_synthesis.cli.main, ["--help"])
    assert result.exit_code == 0
    assert result.output.startswith("Usage: ")

    ctx = Context(axon_synthesis.cli.main)

    for command in axon_synthesis.cli.main.list_commands(ctx):
        result = cli_runner.invoke(axon_synthesis.cli.main, [command, "--help"])
        assert result.exit_code == 0
        assert f"Usage: axon-synthesis {command}" in result.output

    for command in axon_synthesis.cli.validation_group.list_commands(ctx):
        result = cli_runner.invoke(axon_synthesis.cli.main, ["validation", command, "--help"])
        assert result.exit_code == 0
        assert f"Usage: axon-synthesis validation {command}" in result.output


def test_entry_point(script_runner):
    """Test the entry point."""
    ret = script_runner.run("axon-synthesis", "--version")
    assert ret.success
    assert ret.stdout.startswith("axon-synthesis, version ")
    assert ret.stderr == ""



"""Update the properties of the tufts that will be generated later."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging

import pandas as pd

from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import DEFAULT_POPULATION
from axon_synthesis.constants import TARGET_COORDS_COLS
from axon_synthesis.constants import TUFT_COORDS_COLS
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import sublogger


def pick_barcodes(
    terminals,
    edges,
    tuft_properties,
    *,
    rng: SeedType = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Choose a barcode among the ones available."""
    logger = sublogger(logger, __name__)

    if "weight" not in tuft_properties.columns:
        tuft_properties["weight"] = 1

    tuft_properties = tuft_properties[["population_id", "weight", "orientation", "barcode"]].rename(
        columns={"orientation": "tuft_orientation"}
    )

    # Merge edges properties to source and target terminals
    source_terminals = terminals.merge(
        edges.loc[
            edges["source_is_terminal"],
            [
                "source_terminal_id",
                "source_is_terminal",
                "target_is_terminal",
                "section_id",
                "reversed_edge",
            ],
        ].rename(columns={"source_terminal_id": "terminal_id"}),
        on="terminal_id",
    )
    source_terminals["target_is_terminal"] = False
    target_terminals = terminals.merge(
        edges.loc[
            edges["target_is_terminal"],
            [
                "target_terminal_id",
                "source_is_terminal",
                "target_is_terminal",
                "section_id",
                "reversed_edge",
            ],
        ].rename(columns={"target_terminal_id": "terminal_id"}),
        on="terminal_id",
    )
    target_terminals["source_is_terminal"] = False

    edge_terminals = pd.concat(
        [
            source_terminals,
            target_terminals,
        ]
    )

    # Get tuft properties
    edges_with_props = edge_terminals.merge(
        tuft_properties, left_on="target_population_id", right_on="population_id", how="left"
    )

    # Use default barcode for missing populations
    missing_barcode = edges_with_props.loc[edges_with_props["barcode"].isna()]

    if not missing_barcode.empty:
        msg = "No barcode found for the populations %s, the default barcodes are used for them"
        missing_populations = sorted(missing_barcode["target_population_id"].unique().tolist())
        logger.info(msg, missing_populations)
        missing_df = edge_terminals.loc[
            edge_terminals["target_population_id"].isin(missing_populations)
        ].merge(
            tuft_properties.loc[tuft_properties["population_id"] == DEFAULT_POPULATION], how="cross"
        )
        edges_with_props = pd.concat([edges_with_props.dropna(subset="barcode"), missing_df])

    # Format the weight of the tufts
    if "weight" not in edges_with_props.columns:
        if "cluster_weight" in edges_with_props.columns:
            edges_with_props["weight"] = edges_with_props["cluster_weight"]
        else:
            edges_with_props["weight"] = 1
    edges_with_props["weight"] = edges_with_props["weight"].fillna(1)

    # Gather potential barcodes for all terminals
    potential_barcodes = (
        edges_with_props.reset_index()
        .merge(
            edges_with_props.groupby(["terminal_id", "target_population_id"])["weight"]
            .sum()
            .rename("weight_sum"),
            left_on=["terminal_id", "target_population_id"],
            right_index=True,
            how="left",
        )
        .set_index("index")
    )
    potential_barcodes["prob"] = potential_barcodes["weight"] / potential_barcodes["weight_sum"]
    potential_barcodes = potential_barcodes.rename(
        columns={
            TARGET_COORDS_COLS.X: TUFT_COORDS_COLS.X,
            TARGET_COORDS_COLS.Y: TUFT_COORDS_COLS.Y,
            TARGET_COORDS_COLS.Z: TUFT_COORDS_COLS.Z,
        }
    )

    potential_barcodes["use_parent"] = False
    potential_barcodes.loc[
        (potential_barcodes["source_is_terminal"] & ~potential_barcodes["reversed_edge"])
        | (potential_barcodes["target_is_terminal"] & potential_barcodes["reversed_edge"]),
        "use_parent",
    ] = True

    # Pick one barcode for each terminal / target population couple
    return potential_barcodes.groupby(["terminal_id", "target_population_id"]).sample(
        weights="prob", random_state=rng
    )[
        [
            "morphology",
            "axon_id",
            "terminal_id",
            "target_population_id",
            "grafting_section_id",
            *COORDS_COLS,
            "barcode",
            "section_id",
            "target_orientation",
            *TUFT_COORDS_COLS,
            "tuft_orientation",
            "source_is_terminal",
            "target_is_terminal",
            "reversed_edge",
            "use_parent",
        ]
    ]



"""Package to create inputs."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging

from axon_synthesis.atlas import AtlasConfig
from axon_synthesis.inputs import Inputs
from axon_synthesis.inputs.clustering import cluster_morphologies
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import ParallelConfig
from axon_synthesis.white_matter_recipe import WmrConfig

LOGGER = logging.getLogger(__name__)


def create_inputs(
    morphology_dir: FileType,
    output_dir: FileType,
    clustering_parameters: dict,
    wmr_config: WmrConfig | None = None,
    atlas_config: AtlasConfig | None = None,
    neuron_density: float | None = None,
    bouton_density: float | None = None,
    *,
    debug: bool = False,
    rng: SeedType = None,
    parallel_config: ParallelConfig | None = None,
):
    """Create all inputs required to synthesize long-range axons."""
    inputs = Inputs(output_dir, morphology_dir, neuron_density=neuron_density, create=True)

    if atlas_config is not None:
        atlas_config.load_region_map = True

        # Load the Atlas
        inputs.load_atlas(atlas_config)

        # Pre-compute atlas data
        inputs.compute_atlas_region_masks()

    if wmr_config is not None:
        # Process the White Matter Recipe
        inputs.load_wmr(wmr_config)

        # Compute the population and projection probabilities
        inputs.compute_probabilities()

    # Define the tufts and main trunks in input morphologies and compute the properties of the long
    # range trunk and the tufts of each morphology
    inputs.clustering_data = cluster_morphologies(
        inputs.MORPHOLOGY_DIRNAME,
        clustering_parameters,
        inputs.CLUSTERING_DIRNAME,
        atlas=inputs.atlas,
        wmr=inputs.wmr,
        pop_neuron_numbers=inputs.pop_neuron_numbers,
        bouton_density=bouton_density,
        debug=debug,
        rng=rng,
        parallel_config=parallel_config,
    )
    inputs.clustering_data.save()

    # Export the input metadata
    inputs.save_metadata()

    return inputs



"""Clustering from spheres."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components
from scipy.spatial import KDTree

from axon_synthesis.utils import COORDS_COLS

logger = logging.getLogger(__name__)


def compute_clusters(
    config, config_name, axon_id, group_name, group_path, group, output_cols, **_kwargs
):
    """The points must be inside the ball to be merged."""
    # pylint: disable=too-many-locals
    new_terminal_points = []

    # Get the pairs of terminals closer to the given distance
    tree = KDTree(group[COORDS_COLS].to_numpy())
    pairs = tree.query_pairs(config["sphere_radius"])

    # Get the connected components
    adjacency_matrix = np.zeros((len(group), len(group)))
    if pairs:
        adjacency_matrix[tuple(np.array(list(pairs)).T.tolist())] = 1
    np.fill_diagonal(adjacency_matrix, 1)
    graph = csr_matrix(adjacency_matrix)
    _, labels = connected_components(csgraph=graph, directed=False, return_labels=True)

    # Define clusters from these components
    cluster_ids, cluster_sizes = np.unique(labels, return_counts=True)
    big_clusters = cluster_ids[cluster_sizes >= config["min_size"]]
    group_with_label = group.reset_index()
    group_with_label["tuft_id"] = labels
    group_with_label["distance"] = -1.0
    group_with_label["config_name"] = config_name
    clusters = group_with_label.loc[group_with_label["tuft_id"].isin(big_clusters)].groupby(
        "tuft_id",
    )

    # Check clusters
    real_clusters = []
    for cluster_id, cluster in clusters:
        # Check that the potential cluster is a real one (at least 'min_size'
        # points must be close to the center)
        distances, indices = tree.query(
            cluster[COORDS_COLS].mean().to_numpy(),
            k=len(group),
        )
        cluster_mask = np.isin(indices, cluster.index)
        cluster_indices = indices[cluster_mask]
        cluster_distances = distances[cluster_mask]
        if np.count_nonzero(cluster_distances <= config["sphere_radius"]) < config["min_size"]:
            continue

        # Mark the cluster as a real one
        real_clusters.append((cluster_id, cluster))
        group_with_label.loc[cluster_indices, "distance"] = cluster_distances

    # Sort clusters by size
    real_clusters = sorted(real_clusters, key=lambda x: len(x[1]))

    added_clusters = []

    # Merge points from clusters
    for cluster_index, (
        real_cluster_id,
        real_cluster,
    ) in enumerate(real_clusters):
        # Let at least 4 points in the graph
        points_not_clustered = group_with_label.loc[
            ~group_with_label["tuft_id"].isin([i[0] for i in real_clusters[: cluster_index + 1]])
        ]
        if len(points_not_clustered) + cluster_index + 1 <= 3:
            points_in_current_cluster = group_with_label.loc[
                group_with_label["tuft_id"] == real_cluster_id
            ].sort_values("distance", ascending=False)
            removed_indices = points_in_current_cluster.index[
                : max(0, 3 - len(points_not_clustered) - cluster_index)
            ]
            logger.warning(
                "%s: not enough points, removing %s from the cluster %s",
                group_name,
                removed_indices.tolist(),
                real_cluster_id,
            )
            actual_cluster = points_in_current_cluster.loc[
                points_in_current_cluster.index.difference(removed_indices)
            ]

            # Mark the points that will not be merged to keep at least 4 points in the graph
            group_with_label.loc[removed_indices, "tuft_id"] = (
                group_with_label["tuft_id"].max() + np.arange(len(removed_indices)) + 1
            )
        else:
            actual_cluster = real_cluster

        cluster_center = actual_cluster[COORDS_COLS].mean().to_numpy()

        # Add the merged point
        new_terminal_points.append(
            [
                group_name,
                group_path,
                config_name,
                axon_id,
                real_cluster_id,
                len(actual_cluster),
                *cluster_center.tolist(),
            ]
        )
        added_clusters.append(real_cluster_id)

    # Add non merged points
    not_added_mask = ~group_with_label["tuft_id"].isin(added_clusters)
    group_with_label.loc[not_added_mask, "tuft_id"] = sorted(
        set(range(len(added_clusters) + len(group_with_label.loc[not_added_mask]))).difference(
            added_clusters,
        ),
    )
    group_with_label.loc[not_added_mask, "terminal_id"] = group_with_label.loc[
        not_added_mask,
        "tuft_id",
    ]
    group_with_label["size"] = 1
    new_terminal_points.extend(
        group_with_label.loc[not_added_mask, output_cols].to_numpy().tolist()
    )

    return new_terminal_points, group_with_label.set_index("index")["tuft_id"]



"""Create number of neurons in each population."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

LOGGER = logging.getLogger(__name__)


def compute(populations, neuron_density, output_path=None):
    """Compute the number of neurons in each brain region."""
    if output_path is not None:
        output_path = Path(output_path)

        if output_path.exists():
            LOGGER.info(
                "The population neuron numbers are not computed because they already exist in '%s'",
                output_path,
            )
            return None

    populations["pop_neuron_numbers"] = (
        populations["atlas_region_volume"] * neuron_density
    ).astype(int)

    # Format the population numbers
    populations["pop_neuron_numbers"] = populations["pop_neuron_numbers"].clip(lower=2)
    res = populations[
        [
            "pop_raw_name",
            "atlas_region_id",
            "atlas_region_volume",
            "pop_neuron_numbers",
        ]
    ]

    # Export the results
    if output_path is not None:
        output_path.parent.mkdir(parents=True, exist_ok=True)
        res.to_csv(output_path, index=False)

    return res



"""Some utils for graph creation."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging

import numpy as np
import pandas as pd
from neurom.morphmath import angle_between_vectors
from scipy.spatial import Delaunay  # pylint: disable=no-name-in-module
from scipy.spatial import KDTree
from scipy.spatial import Voronoi  # pylint: disable=no-name-in-module

from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import NodeProvider
from axon_synthesis.typing import SeedType

FORCE_2D = False
"""Force the Voronoï and Delaunay calculations to ignore the Z coordinate."""


def add_intermediate_points(
    pts: np.ndarray, ref_coords, min_intermediate_distance, intermediate_number
):
    """Add intermediate points between source points and target points."""
    terms = pts[:, :3] - ref_coords
    term_dists = np.linalg.norm(terms, axis=1)
    if min_intermediate_distance <= 0:
        nb_inter = np.repeat(intermediate_number, len(terms))
    else:
        nb_inter = np.clip(
            term_dists // min_intermediate_distance,
            0,
            intermediate_number,
        )

    if (nb_inter <= 0).all():
        return pts

    inter_pts = []
    for x, y, z, num in np.hstack([terms, np.atleast_2d(nb_inter).T]):
        if num == 0 or np.linalg.norm([x, y, z]) == 0:
            continue
        inter_pts.append(
            (
                num,
                np.hstack(
                    (
                        np.array(
                            [
                                np.linspace(0, x, int(num) + 2)[1:-1],
                                np.linspace(0, y, int(num) + 2)[1:-1],
                                np.linspace(0, z, int(num) + 2)[1:-1],
                            ],
                        ).T
                        + ref_coords,
                        np.ones((int(num), 1)) * NodeProvider.intermediate,
                    )
                ),
            ),
        )
    return np.concatenate([pts, *[i[1] for i in inter_pts]])


def generate_random_points(
    bbox: np.ndarray,
    min_random_point_distance: float,
    rng: SeedType,
    *,
    existing_pts_tree: KDTree | None = None,
    max_tries: int = 10,
):
    """Generate random points in a bounding box and with a min distance."""
    n_fails = 0
    rng = np.random.default_rng(rng)
    new_pts: list[np.ndarray] = []
    while n_fails < max_tries:
        xyz = np.array(
            [
                rng.uniform(bbox[0, 0], bbox[1, 0]),
                rng.uniform(bbox[0, 1], bbox[1, 1]),
                rng.uniform(bbox[0, 2], bbox[1, 2]),
            ],
        )
        if (
            existing_pts_tree is None
            or np.isinf(
                existing_pts_tree.query(
                    xyz,
                    distance_upper_bound=min_random_point_distance,
                    k=2,
                )[0][1],
            )
        ) and (
            len(new_pts) == 0
            or np.linalg.norm(
                xyz - new_pts,
                axis=1,
            ).min()
            > min_random_point_distance
        ):
            new_pts.append(xyz)
            n_fails = 0
        else:
            n_fails += 1
    return new_pts


def add_random_points(
    all_pts: np.ndarray,
    min_random_point_distance: float | None,
    bbox_buffer: float,
    rng: SeedType,
    *,
    max_tries: int = 10,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Add random points in the bounding box of the given points."""
    if min_random_point_distance is not None:
        all_xyz = all_pts[:, :3]
        bbox = np.vstack([all_xyz.min(axis=0), all_xyz.max(axis=0)])
        bbox[0] -= bbox_buffer
        bbox[1] += bbox_buffer
        tree = KDTree(all_xyz)
        new_pts = generate_random_points(
            bbox,
            min_random_point_distance,
            rng=rng,
            existing_pts_tree=tree,
            max_tries=max_tries,
        )

        if new_pts:
            if logger is not None:
                logger.debug("Random points added: %s", len(new_pts))
            all_pts = np.concatenate(
                [
                    all_pts,
                    np.hstack(
                        (np.array(new_pts), np.ones((len(new_pts), 1)) * NodeProvider.random)
                    ),
                ]
            )
        elif logger is not None:
            logger.warning(
                (
                    "Could not add random points! The current state is the following: "
                    "bbox=%s ; nb_pts=%s ; min distance=%s"
                ),
                bbox,
                len(all_pts),
                min_random_point_distance,
            )
    return all_pts


def add_bounding_box_pts(all_pts: np.ndarray):
    """Add points of the bbox."""
    all_xyz = all_pts[:, :3]
    bbox = np.array([all_xyz.min(axis=0), all_xyz.max(axis=0)])
    bbox_pts = np.hstack(
        (
            np.array(np.meshgrid(*np.array(bbox).T)).T.reshape((8, 3)),
            np.ones((8, 1)) * NodeProvider.bbox,
        )
    )
    new_all_pts = np.concatenate([all_pts, bbox_pts])
    return new_all_pts[np.sort(np.unique(new_all_pts, axis=0, return_index=True)[1])]


def add_voronoi_points(
    all_pts: np.ndarray,
    voronoi_steps: int,
    initial_bbox: np.ndarray | None = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Add Voronoi points between the given points."""
    if len(all_pts) < 5:
        return all_pts
    for _ in range(voronoi_steps):
        if FORCE_2D:
            vor = Voronoi(all_pts[:, [0, 1]], qhull_options="QJ")
            step_pts = np.hstack([vor.vertices, np.zeros((len(vor.vertices), 1))])
        else:
            vor = Voronoi(all_pts[:, :3], qhull_options="QJ")
            step_pts = vor.vertices
        if initial_bbox is not None:
            step_pts = step_pts[
                np.all((step_pts >= initial_bbox[0]) & (step_pts <= initial_bbox[1]), axis=1)
            ]
        new_pts = np.hstack([step_pts, np.ones((len(step_pts), 1)) * NodeProvider.Voronoi])
        all_pts = np.concatenate([all_pts, new_pts])  # pylint: disable=no-member
    if logger is not None:
        logger.debug("Added %s Voronoï points using %s steps", len(new_pts), voronoi_steps)
    return all_pts


def drop_close_points(
    all_points_df: pd.DataFrame,
    duplicate_precision: float,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Drop points that are closer to a given distance."""
    tree = KDTree(all_points_df[COORDS_COLS])
    close_pts = tree.query_pairs(duplicate_precision)

    if not close_pts:
        return all_points_df

    to_drop = set()
    for a, b in close_pts:
        label_a = all_points_df.index[a]
        label_b = all_points_df.index[b]
        if label_a not in to_drop and label_b not in to_drop:
            if all_points_df.loc[label_a, "is_terminal"]:
                to_drop.add(label_b)
            else:
                to_drop.add(label_a)

    if logger is not None:
        logger.debug(
            "Dropped %s close points using %s precision", len(to_drop), duplicate_precision
        )
    return all_points_df.drop(list(to_drop))


def drop_outside_points(all_points_df, ref_pts=None, bbox=None):
    """Remove points outside the bounding box of reference points or brain regions."""
    if bbox is not None:
        outside_pts = all_points_df.loc[
            ((all_points_df[COORDS_COLS] < bbox[0]).any(axis=1))
            | ((all_points_df[COORDS_COLS] > bbox[1]).any(axis=1))
        ]
        all_points_df = all_points_df.drop(outside_pts.index)

    if ref_pts is not None:
        ref_xyz = ref_pts[:, :3]
        min_pts = ref_xyz.min(axis=0)
        max_pts = ref_xyz.max(axis=0)
        diff = (max_pts - min_pts) * 0.1
        min_pts -= diff
        max_pts += diff
        outside_pts = all_points_df.loc[
            ((all_points_df[COORDS_COLS] < min_pts).any(axis=1))
            | ((all_points_df[COORDS_COLS] > max_pts).any(axis=1))
        ]
        all_points_df = all_points_df.drop(outside_pts.index)

    return all_points_df


def create_edges(all_points, from_coord_cols, to_coord_cols):
    """Create undirected edges from the Delaunay triangulation of the given points.

    .. note::
        The source-target order has no meaning.
    """
    if len(all_points) < 5:
        msg = (
            "Not enough points to create edges, at least 5 points are needed but only "
            f"{len(all_points)} found"
        )
        raise RuntimeError(msg)

    if FORCE_2D:
        tri = Delaunay(all_points[["x", "y"]], qhull_options="QJ")
    else:
        tri = Delaunay(all_points, qhull_options="QJ")

    # Find all unique edges from the triangulation
    unique_edges = np.unique(
        np.apply_along_axis(
            np.sort,
            1,
            np.vstack(
                # pylint: disable=no-member
                np.stack((tri.simplices, np.roll(tri.simplices, -1, axis=1)), axis=2),
            ),
        ),
        axis=0,
    )

    edges_df = pd.DataFrame(
        {
            "from": unique_edges[:, 0],
            "to": unique_edges[:, 1],
        },
    )

    # Add coordinates and compute base weights equal to the lengths
    edges_df[from_coord_cols] = all_points.loc[edges_df["from"]].to_numpy()
    edges_df[to_coord_cols] = all_points.loc[edges_df["to"]].to_numpy()
    edges_df["length"] = np.linalg.norm(
        edges_df[from_coord_cols].to_numpy() - edges_df[to_coord_cols].to_numpy(),
        axis=1,
    )

    return edges_df, tri


def add_terminal_penalty(edges, nodes):
    """Add penalty to edges to ensure the Steiner algorithm don't connect terminals directly."""
    # Compute penalty
    penalty = edges["weight"].max() + edges["weight"].mean()

    # Get terminal edges
    terminal_edges = edges[["from", "to"]].isin(
        nodes.loc[nodes["is_terminal"], "id"].to_numpy(),
    )

    # Add the penalty
    edges_are_terminals = edges.join(terminal_edges, rsuffix="_is_terminal")
    from_to_all_terminals = edges_are_terminals.groupby("from")[
        ["from_is_terminal", "to_is_terminal"]
    ].all()

    edges_are_terminals = edges_are_terminals.join(
        from_to_all_terminals["from_is_terminal"].rename("from_all_terminals"),
        on="from",
    )
    edges_are_terminals = edges_are_terminals.join(
        from_to_all_terminals["to_is_terminal"].rename("to_all_terminals"),
        on="to",
    )
    edges.loc[
        (edges_are_terminals[["from_is_terminal", "to_is_terminal"]].all(axis=1))
        & (~edges_are_terminals[["from_all_terminals", "to_all_terminals"]].all(axis=1)),
        "weight",
    ] += penalty


def add_orientation_penalty(
    edges_df,
    from_coord_cols,
    to_coord_cols,
    soma_center_coords,
    orientation_penalty_exponent,
    amplitude,
):
    """Add penalty to edges according to their orientation."""
    # atlas_orientations = atlas.load_data("orientation", cls=OrientationField)
    # from_orientations = atlas_orientations.lookup(edges_df[from_coord_cols].to_numpy())
    # to_orientations = atlas_orientations.lookup(edges_df[to_coord_cols].to_numpy())

    # # Compare the two rotation matrices
    # transposed_orientations = np.transpose(from_orientations, axes=[0, 2, 1])
    # dot_prod = np.einsum("...ij,...jk->...ik", transposed_orientations, to_orientations)

    # # The trace of the dot product is equal to the cosine of the angle between the two matrices
    # # and we want to take the cosine of the absolute value of the angle, so we can simplify.
    # penalty = np.abs((np.trace(dot_prod, axis1=1, axis2=2) - 1) * 0.5)

    vectors = edges_df[to_coord_cols].to_numpy() - edges_df[from_coord_cols].to_numpy()
    origin_to_mid_vectors = (
        0.5 * (edges_df[to_coord_cols].to_numpy() + edges_df[from_coord_cols].to_numpy())
        - soma_center_coords
    )
    data = np.stack([origin_to_mid_vectors, vectors], axis=1)

    edge_angles = np.array([angle_between_vectors(i[0], i[1]) for i in data.tolist()])
    return 1 + amplitude * np.power(
        np.clip(np.sin(edge_angles), 1e-3, 1 - 1e-3),
        orientation_penalty_exponent,
    )


def add_depth_penalty(
    edges_df,
    from_coord_cols,
    to_coord_cols,
    depths,
    sigma,
    amplitude,
):
    """Add penalty to edges according to their depth in the Atlas."""
    from_depths = np.nan_to_num(depths.lookup(edges_df[from_coord_cols].to_numpy()))
    to_depths = np.nan_to_num(depths.lookup(edges_df[to_coord_cols].to_numpy()))

    relative_delta = np.clip(np.abs(from_depths - to_depths) / (edges_df["length"]), 0, 1)

    return (1 + amplitude * (1 - np.exp(-relative_delta / sigma))).to_numpy()


def add_preferred_reward(
    edges_df,
    from_coord_cols,
    to_coord_cols,
    preferred_region_tree,
    sigma,
    amplitude,
):
    """Add rewards to edges depending on their distance to the preferred points."""
    from_distances, _ = preferred_region_tree.query(edges_df[from_coord_cols].to_numpy())
    to_distances, _ = preferred_region_tree.query(edges_df[to_coord_cols].to_numpy())

    # TODO: For now we just take the mean of the distance between the start point to the closest
    # preferred point and between the end point to the closest preferred point, which is not
    # accurate.
    return 1 + amplitude * (1 - np.exp(-0.5 * (from_distances + to_distances) / sigma))



"""Define the class to store the inputs."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import json
import logging
from importlib import resources
from pathlib import Path
from typing import ClassVar

import h5py
import pandas as pd
from neurots.validator import ValidationError
from neurots.validator import validate_neuron_distribs
from neurots.validator import validate_neuron_params

from axon_synthesis.atlas import AtlasConfig
from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.base_path_builder import FILE_SELECTION
from axon_synthesis.base_path_builder import BasePathBuilder
from axon_synthesis.inputs import pop_neuron_numbers
from axon_synthesis.inputs.clustering import Clustering
from axon_synthesis.typing import FileType
from axon_synthesis.typing import Self
from axon_synthesis.utils import merge_json_files
from axon_synthesis.utils import recursive_to_str
from axon_synthesis.white_matter_recipe import WhiteMatterRecipe
from axon_synthesis.white_matter_recipe import WmrConfig

LOGGER = logging.getLogger(__name__)

DEFAULT_TUFT_DISTRIBUTIONS = resources.files("axon_synthesis") / "data/tuft_distributions.json"
DEFAULT_TUFT_PARAMETERS = resources.files("axon_synthesis") / "data/tuft_parameters.json"


class Inputs(BasePathBuilder):
    """Class to store the Inputs."""

    _filenames: ClassVar[dict] = {
        "BRAIN_REGIONS_MASK_FILENAME": "region_masks.h5",
        "CLUSTERING_DIRNAME": "Clustering",
        "METADATA_FILENAME": "metadata.json",
        "POPULATION_NEURON_NUMBERS_FILENAME": "neuron_density.csv",
        "POPULATION_PROBABILITIES_FILENAME": "population_probabilities.csv",
        "POPULATION_TUFT_NUMBER_FILENAME": "tufts_numbers_distribution.json",
        "PROJECTION_PROBABILITIES_FILENAME": "projection_probabilities.csv",
        "TUFT_DISTRIBUTIONS_FILENAME": "tuft_distributions.json",
        "TUFT_PARAMETERS_FILENAME": "tuft_parameters.json",
        "WMR_DIRNAME": "WhiteMatterRecipe",
    }

    def __init__(
        self,
        path: FileType,
        morphology_path: FileType | None = None,
        *,
        pop_probability_path: FileType | None = None,
        proj_probability_path: FileType | None = None,
        brain_regions_mask_path: FileType | None = None,
        neuron_density: float | None = None,
        **kwargs,
    ):
        """Create a new Inputs object.

        Args:
            path: The base path used to build the relative paths.
            morphology_path: The path of the directory containing the input morphologies.
            pop_probability_path: The path to the file containing the population probabilities.
            proj_probability_path: The path to the file containing the projection probabilities.
            brain_regions_mask_path: The path to the file containing the pre-computed brain
                region masks.
            neuron_density: The mean neuron density used to compute the expected total number of
                neurons in target regions.
            **kwargs: The keyword arguments are passed to the base constructor.
        """
        super().__init__(path, **kwargs)

        self._brain_regions_mask_file: h5py.File | None = None
        self._pop_neuron_numbers: pd.DataFrame | None = None
        self._metadata: dict
        self.atlas: AtlasHelper | None = None
        self.clustering_data: Clustering | None = None
        self.neuron_density = neuron_density
        self.pop_probability_path = None
        self.population_probabilities: pd.DataFrame | None = None
        self.proj_probability_path = None
        self.projection_probabilities: pd.DataFrame | None = None
        self.tuft_distributions: dict | None = None
        self.tuft_parameters: dict | None = None
        self.tuft_number: pd.DataFrame | None = None
        self.wmr: WhiteMatterRecipe | None = None

        if self.METADATA_FILENAME.exists():
            self.load_metadata()
        else:
            self._metadata = {
                "clustering": self.CLUSTERING_DIRNAME.relative_to(self.path),
                "neuron_density": self.neuron_density,
                "path": self.path,
                "WMR": self.WMR_DIRNAME.relative_to(self.path),
            }
            if morphology_path is not None:
                self._metadata["morphology_path"] = Path(morphology_path)
            if pop_probability_path is not None:
                self._metadata["population_probabilities"] = Path(pop_probability_path)
            if proj_probability_path is not None:
                self._metadata["projection_probabilities"] = Path(proj_probability_path)
            if brain_regions_mask_path is not None:
                self._metadata["brain_regions_mask"] = Path(brain_regions_mask_path)
            self.metadata_to_attributes()

    @staticmethod
    def _format_metadata(metadata: dict) -> dict:
        # Format metadata elements
        metadata["path"] = Path(metadata["path"])
        metadata["clustering"] = Path(metadata["clustering"])
        metadata["morphology_path"] = Path(metadata["morphology_path"])
        metadata["WMR"] = Path(metadata["WMR"])

        if "atlas" in metadata:
            atlas_config = AtlasConfig.from_dict(metadata["atlas"])
            metadata["atlas"] = atlas_config.to_dict()

        return metadata

    @staticmethod
    def _unformat_metadata(metadata: dict) -> dict:
        return recursive_to_str(metadata)

    def _update_atlas_metadata(self) -> None:
        """Update the metadata with atlas config if possible."""
        if "atlas" not in self._metadata and self.atlas is not None:
            self._metadata["atlas"] = self.atlas.config.to_dict()

    @property
    def metadata(self) -> dict:
        """Return the metadata and automatically update them when an atlas is added."""
        self._update_atlas_metadata()
        return self._metadata

    def save_metadata(self):
        """Save the metadata."""
        with self.METADATA_FILENAME.open("w", encoding="utf-8") as f:
            json.dump(
                self._unformat_metadata(self.metadata),
                f,
                indent=4,
                sort_keys=True,
            )

    def load_metadata(self):
        """Load the metadata."""
        try:
            with self.METADATA_FILENAME.open(encoding="utf-8") as f:
                self._metadata = self._format_metadata(json.load(f))
            self.metadata_to_attributes()
        except Exception as exc:
            msg = "Could not load the inputs"
            raise RuntimeError(msg) from exc

    def metadata_to_attributes(self):
        """Propagate metadata to attributes."""
        self._filenames["CLUSTERING_DIRNAME"] = Path(self.metadata["clustering"])
        self._filenames["WMR_DIRNAME"] = Path(self.metadata["WMR"])
        if "brain_regions_mask" in self.metadata:
            self._filenames["BRAIN_REGIONS_MASK_FILENAME"] = str(
                self.metadata["brain_regions_mask"]
            )
        if "morphology_path" in self.metadata:
            self.MORPHOLOGY_DIRNAME = self.metadata["morphology_path"]
        self.neuron_density = self.metadata["neuron_density"]
        self._reset_attributes()

    def load_atlas(self, atlas_config=None):
        """Load the Atlas."""
        if "atlas" not in self.metadata and atlas_config is None:
            msg = "Did not load the Atlas because no atlas configuration was provided."
            LOGGER.debug(msg)
            return
        if atlas_config is None:
            atlas_config = AtlasConfig.from_dict(self.metadata["atlas"])
        elif isinstance(atlas_config, dict):
            atlas_config = AtlasConfig.from_dict(atlas_config)
        # if not getattr(atlas_config, "load_region_map", True):
        #     LOGGER.debug("The 'load_region_map' attribute is forced to 'True'")
        #     atlas_config.load_region_map = True
        if isinstance(atlas_config, AtlasHelper):
            self.atlas = atlas_config
        else:
            self.atlas = AtlasHelper(atlas_config)
        self._update_atlas_metadata()

    def load_brain_regions_masks(self):
        """Load the brain region masks."""
        if self.BRAIN_REGIONS_MASK_FILENAME.exists():
            self._brain_regions_mask_file = h5py.File(self.BRAIN_REGIONS_MASK_FILENAME)
        elif self.atlas is not None:
            LOGGER.debug(
                (
                    "Computing the brain regions masks because the file '%s' is missing but an "
                    "atlas was provided."
                ),
                self.BRAIN_REGIONS_MASK_FILENAME,
            )
            self.compute_atlas_region_masks()
        else:
            LOGGER.debug(
                (
                    "Did not load the brain region masks because the file %s does not exist and "
                    "no atlas was provided."
                ),
                self.BRAIN_REGIONS_MASK_FILENAME,
            )

    def unload_brain_regions_masks(self):
        """Unload the brain region masks."""
        self._brain_regions_mask_file = None

    @property
    def brain_regions_mask_file(self):
        """Return the brain_regions_mask_file attribute or load if it's None."""
        if self._brain_regions_mask_file is None:
            self.load_brain_regions_masks()
        return self._brain_regions_mask_file

    def load_wmr(self, wmr_config: WmrConfig | None = None):
        """Load the Atlas."""
        self.wmr = WhiteMatterRecipe(self.WMR_DIRNAME, load=False)
        if self.wmr.exists():
            self.wmr.load()
        elif wmr_config is None:
            msg = (
                f"The directory '{self.WMR_DIRNAME}' that should contain the White Matter Recipe "
                "does not exist and no config was provided to load a raw WMR"
            )
            raise FileNotFoundError(msg)
        elif self.atlas is None:
            msg = (
                f"The directory '{self.WMR_DIRNAME}' that should contain the White Matter Recipe "
                "does not exist and the atlas is not loaded yet"
            )
            raise FileNotFoundError(msg)
        else:
            self.wmr.load_from_raw_wmr(
                wmr_config,
                self.atlas,
            )
            self.wmr.save()

    def load_clustering_data(self, file_selection=FILE_SELECTION.REQUIRED_ONLY, **kwargs):
        """Load the Atlas."""
        self.clustering_data = Clustering.load(
            self.CLUSTERING_DIRNAME,
            file_selection,
            **kwargs,
        )

    @property
    def pop_neuron_numbers(self):
        """Load the population numbers."""
        if self._pop_neuron_numbers is None:
            if self.POPULATION_NEURON_NUMBERS_FILENAME.exists():
                self._pop_neuron_numbers = pd.read_csv(self.POPULATION_NEURON_NUMBERS_FILENAME)
            elif self.wmr is not None:
                self._pop_neuron_numbers = pop_neuron_numbers.compute(
                    self.wmr.populations,
                    self.neuron_density,
                    self.POPULATION_NEURON_NUMBERS_FILENAME,
                )
            else:
                return pd.DataFrame(
                    columns=[
                        "pop_raw_name",
                        "atlas_region_id",
                        "atlas_region_volume",
                        "pop_neuron_numbers",
                    ]
                )
        return self._pop_neuron_numbers

    def load_probabilities(self, **kwargs):
        """Load the population and projection probabilities."""
        self.update_from_dict(
            "population_probabilities_file", "POPULATION_PROBABILITIES_FILENAME", kwargs
        )
        if self.POPULATION_PROBABILITIES_FILENAME.exists():
            self.population_probabilities = pd.read_csv(self.POPULATION_PROBABILITIES_FILENAME)
            self.population_probabilities = self.population_probabilities.astype(
                {
                    "probability": float,
                    **{
                        col: str
                        for col in self.population_probabilities.columns
                        if col.endswith("population_id")
                    },
                }
            )

        self.update_from_dict(
            "projection_probabilities_file", "PROJECTION_PROBABILITIES_FILENAME", kwargs
        )
        if self.PROJECTION_PROBABILITIES_FILENAME.exists():
            self.projection_probabilities = pd.read_csv(self.PROJECTION_PROBABILITIES_FILENAME)
            self.projection_probabilities = self.projection_probabilities.astype(
                {
                    "probability": float,
                    **{
                        col: str
                        for col in self.projection_probabilities.columns
                        if col.endswith("population_id")
                    },
                }
            )

        self.update_from_dict(
            "population_tuft_number_file", "POPULATION_TUFT_NUMBER_FILENAME", kwargs
        )
        if self.POPULATION_TUFT_NUMBER_FILENAME.exists():
            with self.POPULATION_TUFT_NUMBER_FILENAME.open(mode="r", encoding="utf-8") as f:
                number = pd.DataFrame(json.load(f)).set_index("target_population_id")
            if (number["mean_tuft_number"] <= 0).any():
                LOGGER.error(
                    (
                        "The number of tuft must be greater than 0 for all populations but it's "
                        "not the case for the following ones: %s"
                    ),
                    number.loc[number["mean_tuft_number"] < 0].index.tolist(),
                )
            self.tuft_number = number

    def load_tuft_params_and_distrs(self, **kwargs):
        """Load and validate the parameters and distributions used to generate the tufts."""
        self.update_from_dict("tuft_parameters_file", "TUFT_PARAMETERS_FILENAME", kwargs)
        parameters = merge_json_files(DEFAULT_TUFT_PARAMETERS, self.TUFT_PARAMETERS_FILENAME)
        try:
            validate_neuron_params(parameters)
        except ValidationError as exc:
            msg = "The given tuft parameters are not valid"
            raise ValidationError(msg) from exc
        self.tuft_parameters = parameters

        self.update_from_dict("tuft_distributions_file", "TUFT_DISTRIBUTIONS_FILENAME", kwargs)
        distributions = merge_json_files(
            DEFAULT_TUFT_DISTRIBUTIONS, self.TUFT_DISTRIBUTIONS_FILENAME
        )
        try:
            validate_neuron_distribs(distributions)
        except ValidationError as exc:
            msg = "The given tuft distributions are not valid"
            raise ValidationError(msg) from exc
        self.tuft_distributions = distributions

    @classmethod
    def load(
        cls,
        path: FileType | None,
        atlas_config: AtlasConfig | AtlasHelper | None = None,
        **kwargs: dict,
    ) -> Self:
        """Load all the inputs from the given path."""
        if path is None:
            missing_keys = {
                "projection_probabilities_file",
                "population_probabilities_file",
                "trunk_properties_file",
                "tuft_properties_file",
            }.difference(kwargs.keys())
            if missing_keys:
                msg = (
                    "The following keys are missing in kwargs because the 'path' is not "
                    f"provided: {sorted(missing_keys)}"
                )
                raise RuntimeError(msg)
        obj = cls(path or "")
        obj.load_atlas(atlas_config)
        obj.load_clustering_data(**kwargs)
        obj.load_brain_regions_masks()
        obj.load_probabilities(**kwargs)
        obj.load_tuft_params_and_distrs(**kwargs)
        if obj.pop_neuron_numbers is None:
            msg = "Could not load or build the population numbers in target regions"
            raise RuntimeError(msg)
        return obj

    def export_probabilities(self):
        """Export the computed probabilities."""
        # Export the population probabilities
        if self.population_probabilities is not None:
            self.population_probabilities.to_csv(
                self.POPULATION_PROBABILITIES_FILENAME, index=False
            )

        # Export the projection probabilities
        if self.projection_probabilities is not None:
            self.projection_probabilities.to_csv(
                self.PROJECTION_PROBABILITIES_FILENAME, index=False
            )

    def compute_probabilities(self, source="WMR", *, export=True):
        """Compute projection probabilities."""
        if (
            self.POPULATION_PROBABILITIES_FILENAME.exists()
            and self.PROJECTION_PROBABILITIES_FILENAME.exists()
        ):
            LOGGER.info(
                "The population and projection probabilities are not computed because they already "
                "exists in '%s' and '%s'",
                self.POPULATION_PROBABILITIES_FILENAME,
                self.PROJECTION_PROBABILITIES_FILENAME,
            )
            return None

        # TODO: For now we support only the WMR but other methods may come later.
        if source == "WMR" and self.wmr is not None and self.atlas is not None:
            (
                self.population_probabilities,
                self.projection_probabilities,
            ) = self.wmr.compute_probabilities(self.atlas)
        else:
            msg = f"The value '{source}' is not known."
            raise ValueError(msg)

        if export:
            self.export_probabilities()
        return self.population_probabilities, self.projection_probabilities

    def compute_atlas_region_masks(self):
        """Compute all region masks of the Atlas."""
        if self.atlas is None:
            msg = "The Atlas must be loaded before computing the region masks."
            raise RuntimeError(msg)
        self.atlas.compute_region_masks(self.BRAIN_REGIONS_MASK_FILENAME)
        self._brain_regions_mask_file = h5py.File(self.BRAIN_REGIONS_MASK_FILENAME)



"""Clustering from barcodes."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

# mypy: ignore-errors
# import json
from collections import defaultdict
from itertools import pairwise

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from morphio import IterType
from scipy import stats
from tmd.io.io import load_neuron_from_morphio
from tmd.Topology.analysis import barcode_bin_centers
from tmd.Topology.analysis import histogram_horizontal
from tmd.Topology.analysis import histogram_stepped
from tmd.Topology.methods import tree_to_property_barcode
from tmd.view.plot import barcode as plot_barcode

from axon_synthesis.utils import COORDS_COLS


def barcode_mins(barcode, nb_bins=100, threshold=0.1):
    """Compute min values of a barcode."""
    bin_centers, data = barcode_bin_centers(barcode, num_bins=nb_bins)

    # Gaussian kernel to smooth distribution of bars
    kde = stats.gaussian_kde(data)
    minimas = []

    # Compute first and second derivatives
    der1 = np.gradient(kde(bin_centers))
    der2 = np.gradient(der1)

    # Compute minimas of distribution
    while len(minimas) == 0:
        minimas = np.where(abs(der1) < threshold * np.max(abs(der1)))[0]
        minimas = minimas[der2[minimas] > 0]
        threshold *= 2.0  # if threshold was too small, increase and retry

    def _get_min_indices(mins, der) -> np.ndarray:
        # Compute where the derivative crosses the X axis
        der_zero_intervals = np.append(
            np.insert(
                np.where(np.diff(np.sign(der)))[0],
                0,
                -len(der) * 2,
            ),
            len(der) * 2,
        )

        # Find in which interval is each value
        zero_interval_indices = np.digitize(mins, der_zero_intervals)
        _tmp = pd.DataFrame(
            {
                "min_indices": mins,
                "interval_idx": zero_interval_indices,
            },
        ).groupby("interval_idx")

        # Get the median value
        return _tmp.quantile(interpolation="higher")["min_indices"].astype(int).to_numpy()

    # Keep one minimum per der1 and der2 interval
    min_indices = _get_min_indices(minimas, der1)
    min_indices = _get_min_indices(min_indices, der2)
    min_positions = bin_centers[min_indices]

    return min_indices, min_positions, bin_centers, der1, der2


def compute_clusters(
    config_str,  # noqa: ARG001
    morph,
    axon,  # noqa: ARG001
    axon_id,  # noqa: ARG001
    group_name,  # noqa: ARG001
    group_path,
    group,
    output_cols,  # noqa: ARG001
    *,
    debug=False,
    **_kwargs,
):
    """The points must be inside the ball to be merged."""
    # pylint: disable=too-many-locals
    # pylint: disable=unused-argument
    msg = "This mode is not implemented yet."
    raise NotImplementedError(msg)
    # pylint: disable=unreachable
    # pylint: disable=unused-variable
    new_terminal_points = []

    # config_str = json.dumps(config)

    # Get the complete morphology
    # morph = load_morphology(group_name)
    soma_center = morph.soma.center
    # axons = [i for i in neuron.neurites if i.type == NeuriteType.axon]

    neuron = load_neuron_from_morphio(group_path)
    origin = morph.soma.center
    nb_bins = 100

    for neuron_axon in neuron.axon:
        barcode, bars_to_points = tree_to_property_barcode(
            neuron_axon,
            # lambda tree: tree.get_point_path_distances(),
            lambda tree: tree.get_point_radial_distances(point=origin),
        )

        min_indices, min_positions, bin_centers, der1, der2 = barcode_mins(
            barcode,
            nb_bins,
        )

        # Plot
        if debug:
            fig, (ax_barcode, ax_hist, ax_der) = plt.subplots(1, 3, figsize=(12, 9))

            # Plot barcode
            plt.sca(ax_barcode)
            plot_barcode(barcode, new_fig=False)
            ax_barcode.vlines(
                min_positions,
                0,
                ax_barcode.get_ylim()[1],
                color="red",
                linestyle="--",
                label="In-between tufts",
            )

            # Plot histograms
            plt.sca(ax_hist)
            hist_data_horizontal = histogram_horizontal(barcode, num_bins=nb_bins)
            ax_hist.plot(
                hist_data_horizontal[0][:-1],
                hist_data_horizontal[1],
                color="chocolate",
                alpha=0.7,
                label="Histogram horizontal",
            )
            hist_data_stepped = histogram_stepped(barcode)
            ax_hist.plot(
                hist_data_stepped[0][:-1],
                hist_data_stepped[1],
                color="blue",
                alpha=0.7,
                label="Histogram stepped",
            )
            ax_hist.vlines(
                min_positions,
                0,
                np.max(np.concatenate([hist_data_horizontal[1], hist_data_stepped[1]])),
                color="red",
                linestyle="--",
                label="In-between tufts",
            )
            ax_hist.legend()

            # Plot derivatives
            ax_der.plot(bin_centers, der1, color="chocolate", alpha=0.7, label="1st derivative")
            ax_der.plot(bin_centers, der2, color="blue", alpha=0.7, label="2nd derivative")
            ax_der.vlines(
                min_positions,
                np.min(np.concatenate([der1, der2])),
                np.max(np.concatenate([der1, der2])),
                color="red",
                linestyle="--",
                label="In-between tufts",
            )
            ax_der.legend()

            plt.show()

        group["radial_dist"] = np.linalg.norm(group[COORDS_COLS] - soma_center, axis=1)
        min_positions = np.append(np.insert(min_positions, 0, 0), group["radial_dist"].max() + 1)
        terminal_intervals = np.digitize(group["radial_dist"], min_positions)
        cluster_ids = []

        for num_interval, interval in enumerate(pairwise(min_positions)):
            cluster_terminals = group.loc[terminal_intervals == num_interval + 1]
            terminal_parents = defaultdict(list)
            crossing_sections = set()
            for term_sec in cluster_terminals["section_id"].to_numpy():
                for sec in morph.section(term_sec).iter(IterType.upstream):
                    if np.linalg.norm(sec.points[-1] - soma_center) < interval[0]:
                        break
                    if np.linalg.norm(sec.points[0] - soma_center) <= interval[0]:
                        crossing_sections.add(sec.id)
                    terminal_parents[sec.id].append(term_sec)
            if not crossing_sections:
                crossing_sections.add(min(terminal_parents.keys()))

            # raise NotImplementedError("This mode is not implemented yet.")
            # for sec in crossing_sections:
            #     print(sec)

    return new_terminal_points, cluster_ids



"""Extract the terminal points of a morphology so that a Steiner Tree can be computed on them."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging

import pandas as pd
from bluepyparallel import evaluate
from bluepyparallel import init_parallel_factory
from dask.distributed import LocalCluster

from axon_synthesis.typing import FileType
from axon_synthesis.utils import COORDS_COLS
from axon_synthesis.utils import ParallelConfig
from axon_synthesis.utils import disable_distributed_loggers
from axon_synthesis.utils import get_axons
from axon_synthesis.utils import get_morphology_paths
from axon_synthesis.utils import load_morphology

LOGGER = logging.getLogger(__name__)


def process_morph(
    morph_path: FileType, morph_name: str
) -> list[tuple[str, int, int, int, float, float, float]]:
    """Extract the terminal points from a morphology."""
    morph_path_str = str(morph_path)
    morph = load_morphology(morph_path)
    pts = []
    axons = get_axons(morph)

    nb_axons = len(axons)
    log_func = LOGGER.warning if nb_axons == 0 else LOGGER.info
    log_func("%s: %s axon%s found", morph_name, nb_axons, "s" if nb_axons > 1 else "")

    for axon_id, axon in enumerate(axons):
        # Add root point
        pts.append(
            (
                morph_name,
                morph_path_str,
                axon_id,
                0,
                axon.root_node.id,
                *axon.root_node.points[0][:3].tolist(),
            ),
        )

        # Add terminal points
        terminal_id = 1
        for section in axon.iter_sections():
            if not section.children:
                pts.append(
                    (
                        morph_name,
                        morph_path_str,
                        axon_id,
                        terminal_id,
                        section.id,
                        *section.points[-1][:3].tolist(),
                    ),
                )
                terminal_id += 1

    return pts


def _wrapper(data: dict) -> dict:
    """Wrap process_morph() for parallel computation."""
    return {"res": process_morph(data["morph_path"], data["morph_name"])}


def process_morphologies(
    morph_dir: FileType, parallel_config: ParallelConfig | None = None
) -> pd.DataFrame:
    """Extract terminals from all the morphologies in the given directory."""
    if parallel_config is None:
        parallel_config = ParallelConfig()

    morphologies = get_morphology_paths(morph_dir)

    with disable_distributed_loggers():
        if parallel_config.nb_processes > 1:
            LOGGER.info("Start parallel computation using %s workers", parallel_config.nb_processes)
            cluster = LocalCluster(n_workers=parallel_config.nb_processes, timeout="60s")
            parallel_factory = init_parallel_factory("dask_dataframe", address=cluster)
        else:
            LOGGER.info("Start computation")
            parallel_factory = init_parallel_factory(None)

        # Extract terminals of each morphology
        results = evaluate(
            morphologies,
            _wrapper,
            [
                ["res", None],
            ],
            parallel_factory=parallel_factory,
        )

        # Close the Dask cluster if opened
        if parallel_config.nb_processes > 1:
            parallel_factory.shutdown()
            cluster.close()

    results = results.loc[results["res"].apply(len) > 0]
    final_results = results["res"].explode().apply(pd.Series)
    final_results.columns = ["morph_file", "axon_id", "terminal_id", "section_id", *COORDS_COLS]
    if final_results.empty:
        msg = "No morphology with axon found"
        raise RuntimeError(msg)
    return final_results



"""Validation workflow that mimics inputs morphologies."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
import re
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
from attrs import evolve
from morph_tool.converter import convert
from neurom.geom.transform import Translation
from voxcell.cell_collection import CellCollection

from axon_synthesis.constants import ATLAS_COORDS_COLS
from axon_synthesis.constants import AXON_GRAFTING_POINT_HDF_GROUP
from axon_synthesis.constants import COMMON_ANCESTOR_COORDS_COLS
from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import DEFAULT_OUTPUT_PATH
from axon_synthesis.constants import DEFAULT_POPULATION
from axon_synthesis.constants import TARGET_COORDS_COLS
from axon_synthesis.inputs import Inputs
from axon_synthesis.inputs.create import create_inputs
from axon_synthesis.synthesis import ParallelConfig
from axon_synthesis.synthesis import SynthesisConfig
from axon_synthesis.synthesis import synthesize_axons
from axon_synthesis.synthesis.main_trunk.create_graph import CreateGraphConfig
from axon_synthesis.synthesis.main_trunk.post_process import PostProcessConfig
from axon_synthesis.synthesis.outputs import OutputConfig
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import disable_loggers
from axon_synthesis.utils import get_nested_morphology_paths
from axon_synthesis.utils import load_morphology
from axon_synthesis.utils import parallel_evaluator
from axon_synthesis.utils import seed_from_name
from axon_synthesis.utils import setup_logger
from axon_synthesis.validation.atlas_utils import morph_atlas
from axon_synthesis.validation.utils import copy_morph_to_tmp_dir

LOGGER = logging.getLogger(__name__)


def create_cell_collection(
    morphology_dir,
    output_path: FileType | None = None,
    convert_to: FileType | None = None,
    *,
    recenter: bool = True,
) -> CellCollection:
    """Create a CellCollection object from a directory containing morphologies."""
    morphology_dir = Path(morphology_dir)
    raw_morph_files = get_nested_morphology_paths(morphology_dir)

    centers = np.zeros((len(raw_morph_files), 3), dtype=float)
    registered_centers = np.zeros((len(raw_morph_files), 3), dtype=float)
    if convert_to is not None:
        convert_to = Path(convert_to)
        convert_to.mkdir(parents=True, exist_ok=True)
        morph_files = []
        for num, file in enumerate(raw_morph_files):
            converted_file = (convert_to / file.relative_to(morphology_dir)).with_suffix(".h5")
            converted_file.parent.mkdir(parents=True, exist_ok=True)
            morph = load_morphology(file)
            registered_centers[num] = morph.soma.center
            if recenter:
                morph = morph.transform(Translation(-morph.soma.center))
            with disable_loggers("morph_tool.converter"):
                convert(morph, converted_file, nrn_order=True, sanitize=True)
            morph_files.append(converted_file)
    else:
        morph_files = raw_morph_files

    morph_names = [str(i.relative_to(morphology_dir).with_suffix("")) for i in raw_morph_files]
    raw_morph_files = [str(i) for i in raw_morph_files]
    morph_files = [str(i) for i in morph_files]
    if not morph_files:
        msg = f"No morphology file found in '{morphology_dir}'"
        raise RuntimeError(msg)
    cells_df = pd.DataFrame(
        {"morphology": morph_names, "raw_file": raw_morph_files, "morph_file": morph_files}
    )
    cells_df["mtype"] = DEFAULT_POPULATION
    cells_df["region"] = DEFAULT_POPULATION
    if recenter:
        cells_df[COORDS_COLS] = centers
    else:
        cells_df[COORDS_COLS] = registered_centers
    cells_df[ATLAS_COORDS_COLS] = registered_centers
    cells_df["orientation"] = [np.eye(3)] * len(cells_df)
    cells_df = cells_df.sort_values("morphology", ignore_index=True)
    cells_df.loc[:, "seed"] = cells_df["morphology"].apply(seed_from_name)
    cells_df.index += 1
    cells = CellCollection.from_dataframe(cells_df)
    if output_path is not None:
        cells.save(output_path)
    LOGGER.info("Found %s morphologies in %s", len(cells_df), morphology_dir)
    return cells


def workflow_col_name(name):
    """Create a column name based on a workflow name."""
    return f"synthesized_{name}_path"


def create_probabilities(cells_df, tuft_properties):
    """Create the population and projection probabilities."""
    # Create population IDs with one unique population for each morphology-axon couple
    tuft_properties["population_id"] = tuft_properties.apply(
        lambda row: f"{row['morphology']}_{row['axon_id']}", axis=1
    )

    # Create population probabilities with all probs = 1
    population_probabilities = (
        tuft_properties[["population_id"]].drop_duplicates().reset_index(drop=True)
    )

    population_probabilities["brain_region_id"] = population_probabilities["population_id"]
    population_probabilities["probability"] = 1

    # Create projection probabilities with all probs = 1
    projection_probabilities = (
        (
            cells_df[["morphology", "morph_file"]].merge(
                tuft_properties[
                    [
                        "morphology",
                        "axon_id",
                        "tuft_id",
                        "population_id",
                        *COMMON_ANCESTOR_COORDS_COLS,
                    ]
                ],
                on="morphology",
                how="left",
            )
        )
        .dropna(subset="tuft_id")
        .rename(columns={"population_id": "source_population_id"})
    )
    projection_probabilities["source_brain_region_id"] = projection_probabilities[
        "source_population_id"
    ].copy()
    projection_probabilities["target_population_id"] = (
        projection_probabilities["source_population_id"]
        + "-"
        + projection_probabilities["tuft_id"].astype(str)
    )
    projection_probabilities["target_brain_region_id"] = projection_probabilities[
        "target_population_id"
    ]
    projection_probabilities[TARGET_COORDS_COLS] = projection_probabilities[
        COMMON_ANCESTOR_COORDS_COLS
    ]
    projection_probabilities["probability"] = 1

    return population_probabilities, projection_probabilities


def update_cells(cells_df, projection_probabilities):
    """Update the cell collection according to the computed projection probabilities."""
    # Copy the dummy source brain region ID to cells_df
    cells_df["source_brain_region_id"] = cells_df.merge(
        projection_probabilities[["morphology", "source_brain_region_id"]].drop_duplicates(
            subset=["morphology"]
        ),
        on="morphology",
        how="left",
    )["source_brain_region_id"].to_numpy()

    # Drop morphologies with missing brain region IDs
    missing_brain_regions = cells_df.loc[cells_df["source_brain_region_id"].isna()]
    if not missing_brain_regions.empty:
        LOGGER.warning(
            "Could not find the source brain region of the following morphologies: %s",
            ", ".join(missing_brain_regions["morphology"].sort_values().tolist()),
        )
        cells_df.drop(missing_brain_regions.index, inplace=True)
        cells_df.reset_index(drop=True, inplace=True)
        cells_df.index += 1


def mimic_preferred_regions_workflow(
    morph,
    voxel_dimensions,
    synthesis_config,
    create_graph_config,
    post_process_config,
    output_config,
    rng,
    *,
    keep_tmp_atlas=False,
):
    """Synthesize axons using morphology preferred regions."""
    atlas_tmp_dir, atlas_tmp = morph_atlas(
        morph,
        voxel_dimensions=voxel_dimensions,
        preferred_regions_axon_id=0,  # TODO: Handle multiple axons case
        export=keep_tmp_atlas,
        tmp_dir=output_config.path / "dummy_atlas",
    )
    create_graph_config_tmp = evolve(
        create_graph_config,
        preferred_regions=["dft"],
    )
    res = synthesize_axons(
        synthesis_config,
        atlas_config=atlas_tmp,
        create_graph_config=create_graph_config_tmp,
        post_process_config=post_process_config,
        output_config=output_config,
        rng=rng,
    )
    if not keep_tmp_atlas:
        atlas_tmp_dir.cleanup()
    return res


def run_workflows(
    data,
    workflows,
    synthesis_config,
    create_graph_config,
    post_process_config,
    output_config,
    rng,
    *,
    voxel_dimensions=None,
    keep_tmp_atlas=False,
):
    """Run all the workflows on each morphology."""
    morph_name = data["morphology"]

    # Prepare specific inputs for the current morphology
    tmp_dir, _filename = copy_morph_to_tmp_dir(data["raw_file"])
    synthesis_config_tmp = evolve(
        synthesis_config,
        morphology_dir=tmp_dir.name,
        morphology_data_file=Path(tmp_dir.name) / "circuit.h5",
    )

    # Create a circuit containing only the current morphology
    cells_df = CellCollection.load(synthesis_config.morphology_data_file).as_dataframe()
    cell_row = cells_df.loc[cells_df["morphology"] == morph_name].reset_index(drop=True)
    cell_row.index += 1
    CellCollection.from_dataframe(cell_row).save(synthesis_config_tmp.morphology_data_file)

    synthesized_paths = {}

    # Synthesize each morphology
    for workflow in workflows:
        output_config_tmp = evolve(
            output_config,
            path=output_config.path.parent
            / Path(morph_name).parent
            / (output_config.path.name + f"_{workflow}_{Path(morph_name).name}"),
        )
        if workflow == "basic":
            LOGGER.info("Starting 'basic' workflow for %s", morph_name)
            res = synthesize_axons(
                synthesis_config_tmp,
                create_graph_config=create_graph_config,
                post_process_config=post_process_config,
                output_config=output_config_tmp,
                rng=rng,
            )
        elif workflow == "preferred_regions":
            LOGGER.info("Starting 'preferred regions' workflow for %s", morph_name)
            if voxel_dimensions is None:
                msg = "A 'voxel_dimensions' value is required for this workflow"
                raise ValueError(msg)

            res = mimic_preferred_regions_workflow(
                data["morph_file"],
                voxel_dimensions,
                synthesis_config_tmp,
                create_graph_config=create_graph_config,
                post_process_config=post_process_config,
                output_config=output_config_tmp,
                rng=rng,
                keep_tmp_atlas=keep_tmp_atlas,
            )
        else:
            msg = f"The workflow {workflow} is not known"
            raise ValueError(msg)
        synthesized_paths[workflow] = res["file_path"].iloc[0]

    # Cleanup tmp files
    tmp_dir.cleanup()

    return {workflow_col_name(k): v for k, v in synthesized_paths.items()}


def export_cells_df(cells_df, path):
    """Export cells DF for future debugging."""
    cells_df = cells_df.copy(deep=False)
    cells_df["orientation"] = cells_df["orientation"].apply(np.ndarray.tolist)
    cells_df.reset_index().to_feather(path)


def merge_result_folders(results, output_path):
    """Create a new folder with simlinks to all result sub-folders."""
    # import pdb
    # pdb.set_trace()
    # print(results)
    LOGGER.info("Merging all results into %s", output_path)
    LOGGER.warning("The merging feature is not implemented yet!")
    return results


def mimic_axons(  # noqa: PLR0913
    morphology_dir: FileType,
    clustering_parameters: dict,
    *,
    tuft_parameters_file: FileType | None = None,
    create_graph_config: CreateGraphConfig | None = None,
    post_process_config: PostProcessConfig | None = None,
    output_config: OutputConfig | None = None,
    workflows: list[str] | None = None,
    voxel_dimensions: int | None = None,
    rng: SeedType = None,
    debug: bool = False,
    parallel_config: ParallelConfig | None = None,
    keep_tmp_atlas: bool = False,
    merge_results: bool = True,
):
    """Synthesize mimicking axons."""
    output_config = (
        output_config if output_config is not None else OutputConfig(DEFAULT_OUTPUT_PATH)
    )
    input_dir = output_config.path / "inputs"
    output_config.path = output_config.path / "synthesis"

    if len(clustering_parameters) != 1:
        msg = "The 'clustering_parameters' JSON object should contain exactly 1 entry."
        raise ValueError(msg)

    # Convert the morphologies to h5 and create a CellCollection
    morphology_data_file = input_dir / "circuit.h5"
    converted_morphologies_dir = input_dir / "converted_morphologies"
    cells = create_cell_collection(morphology_dir, morphology_data_file, converted_morphologies_dir)
    cells_df = cells.as_dataframe()

    # Create raw inputs if they don't exist
    inputs = Inputs(input_dir, converted_morphologies_dir)
    if not inputs.CLUSTERING_DIRNAME.exists():
        inputs = create_inputs(
            converted_morphologies_dir,
            input_dir,
            clustering_parameters,
            rng=rng,
            debug=debug,
            parallel_config=parallel_config,
        )
    else:
        inputs.load_clustering_data()

    # Modify the inputs for the mimic workflow

    # Build and export the probabilities
    inputs.population_probabilities, inputs.projection_probabilities = create_probabilities(
        cells_df,
        inputs.clustering_data.tuft_properties,  # type: ignore[union-attr]
    )
    inputs.export_probabilities()

    # Update the cell properties
    update_cells(cells_df, inputs.projection_probabilities)

    # Update tuft properties
    tuft_properties = inputs.clustering_data.tuft_properties.merge(  # type: ignore[union-attr]
        inputs.projection_probabilities[  # type: ignore[index]
            ["morphology", "axon_id", "tuft_id", "target_population_id"]
        ],
        on=["morphology", "axon_id", "tuft_id"],
        how="left",
    )
    inputs.clustering_data.tuft_properties["population_id"] = tuft_properties[  # type: ignore[union-attr, index]
        "target_population_id"
    ].to_numpy()
    inputs.clustering_data.save()  # type: ignore[union-attr]

    # Set the source properties
    cells_df["st_level"] = 0
    export_cells_df(cells_df, output_config.path.parent / "cells_df.feather")
    cells = CellCollection.from_dataframe(cells_df)
    cells.save(morphology_data_file)

    # Create the axon grafting point file
    axon_grafting_points_file = output_config.path.parent / "axon_grafting_points.h5"
    axon_grafting_points = (
        tuft_properties[["morphology", "population_id"]].drop_duplicates().reset_index(drop=True)
    )
    axon_grafting_points["grafting_section_id"] = -1
    axon_grafting_points["source_brain_region_id"] = axon_grafting_points["population_id"]
    axon_grafting_points["rebuilt_existing_axon_id"] = axon_grafting_points.groupby(
        "morphology"
    ).cumcount()
    axon_grafting_points.to_hdf(axon_grafting_points_file, key=AXON_GRAFTING_POINT_HDF_GROUP)

    # Synthesize the axons using the modified inputs
    synthesis_config = SynthesisConfig(
        converted_morphologies_dir,
        morphology_data_file,
        axon_grafting_points_file,
        input_dir,
        tuft_parameters_file=tuft_parameters_file,
        rebuild_existing_axons=True,
    )

    # Choose the workflows to process
    if workflows is None:
        workflows = ["basic"]

    workflow_cols = [workflow_col_name(workflow) for workflow in workflows]

    results = parallel_evaluator(
        cells_df,
        run_workflows,
        parallel_config,
        [[col, None] for col in workflow_cols],
        func_kwargs={
            "workflows": workflows,
            "synthesis_config": synthesis_config,
            "create_graph_config": create_graph_config,
            "post_process_config": post_process_config,
            "output_config": output_config,
            "rng": rng,
            "voxel_dimensions": voxel_dimensions,
            "keep_tmp_atlas": keep_tmp_atlas,
        },
        progress_bar=False,
        startup_func=partial(setup_logger, level=logging.getLevelName(LOGGER.getEffectiveLevel())),
    )

    failed = results[workflow_cols].isna().any(axis=1)
    if failed.any():
        for col in workflow_cols:
            workflow = re.match("synthesized_(.*)_path", col).group(1)  # type: ignore[union-attr]
            LOGGER.error(
                "The following morphologies could not be synthesized for the workflow %s (see the "
                "logs for details): %s",
                workflow,
                sorted(results.loc[results[col].isna(), "morphology"].tolist()),
            )
    LOGGER.info(
        "Synthesized %s morphologies on %s for the following workflows: %s",
        (~failed).sum(),
        len(results),
        ", ".join(workflows),
    )
    if merge_results:
        pass
    #     merge_result_folders(results, output_config.path)
    return results



"""Some plot utils for create graph."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly_helper.neuron_viewer import NeuronBuilder

from axon_synthesis.constants import TARGET_COORDS_COLS
from axon_synthesis.utils import add_camera_sync
from axon_synthesis.utils import build_layout_properties
from axon_synthesis.utils import get_morph_pts


def plot_final_morph(morph, target_points, output_path, initial_morph=None, logger=None):
    """Plot the given morphology.

    If `initial_morph` is not None then the given morphology is also plotted for comparison.
    """
    title = "Final morphology"
    fig_builder = NeuronBuilder(morph, "3d", line_width=4, title=title)
    fig_data = [fig_builder.get_figure()["data"]]
    left_title = "Synthesized morphology"

    if initial_morph is not None:
        fig = make_subplots(
            cols=2,
            specs=[[{"type": "scene"}, {"type": "scene"}]],
            subplot_titles=[left_title, "Initial morphology"],
        )

        if initial_morph.root_sections:
            fig_data.append(
                NeuronBuilder(initial_morph, "3d", line_width=4, title=title).get_figure()["data"]
            )
        else:
            fig_data.append(
                [
                    go.Scatter3d(
                        x=[initial_morph.soma.center[0]],
                        y=[initial_morph.soma.center[1]],
                        z=[initial_morph.soma.center[2]],
                        marker={"color": "black", "size": 4},
                        mode="markers",
                        name="Soma",
                    )
                ]
            )
    else:
        fig = make_subplots(cols=1, specs=[[{"type": "scene"}]], subplot_titles=[left_title])

    for col_num, data in enumerate(fig_data):
        fig.add_traces(data, rows=[1] * len(data), cols=[col_num + 1] * len(data))

    x, y, z = target_points[TARGET_COORDS_COLS].to_numpy().T
    node_trace = go.Scatter3d(
        x=x,
        y=y,
        z=z,
        mode="markers",
        marker={"size": 2, "color": "chocolate"},
        name="Target points",
    )
    fig.add_trace(node_trace, row=1, col=1)

    layout_props = build_layout_properties(
        np.vstack([morph.points, get_morph_pts(initial_morph)]), 0.1
    )

    fig.update_scenes(layout_props)
    fig.update_layout(title=morph.name)

    # Export figure
    Path(output_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(output_path)

    if initial_morph is not None:
        add_camera_sync(output_path)

    if logger is not None:
        logger.info("Exported figure to %s", output_path)


def plot_target_points(morph, source_point, target_points, output_path, logger=None):
    """Plot the source and target points along the given morphology."""
    title = "Initial morphology"

    fig = make_subplots(cols=1, specs=[[{"type": "scene"}]], subplot_titles=[title])

    if morph.root_sections:
        fig.add_traces(NeuronBuilder(morph, "3d", line_width=4, title=title).get_figure()["data"])
    else:
        fig.add_traces(
            go.Scatter3d(
                x=[morph.soma.center[0]],
                y=[morph.soma.center[1]],
                z=[morph.soma.center[2]],
                marker={"color": "black", "size": 4},
                mode="markers",
                name="Soma",
            )
        )

    source_point_trace = go.Scatter3d(
        x=[source_point[0]],
        y=[source_point[1]],
        z=[source_point[2]],
        marker={"color": "purple", "size": 4},
        mode="markers",
        name="Source point",
    )

    target_points_trace = go.Scatter3d(
        x=target_points[:, 0],
        y=target_points[:, 1],
        z=target_points[:, 2],
        marker={"color": "chocolate", "size": 3},
        mode="markers",
        name="Target points",
    )

    fig.add_trace(source_point_trace)
    fig.add_trace(target_points_trace)

    layout_props = build_layout_properties(
        np.concatenate(
            [
                (
                    morph.points
                    if len(morph.root_sections) > 0
                    else np.atleast_2d(morph.soma.center)
                )[:, :3],
                [source_point],
                target_points,
            ]
        ),
        0.1,
    )

    fig.update_scenes(layout_props)
    fig.update_layout(title=morph.name)

    # Export figure
    Path(output_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(output_path)

    if logger is not None:
        logger.debug("Exported figure to %s", output_path)



"""Module to create custom atlases."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import json
import logging
from collections import OrderedDict
from pathlib import Path

import numpy as np
from attrs import evolve
from brainbuilder.app.atlases import _dump_atlases  # type: ignore[import-untyped]
from brainbuilder.app.atlases import _normalize_hierarchy
from neurom import COLS
from neurom.core import Morphology
from voxcell import RegionMap
from voxcell import VoxelData

from axon_synthesis.atlas import AtlasConfig
from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.utils import CleanableDirectory
from axon_synthesis.utils import compute_bbox
from axon_synthesis.utils import get_axons
from axon_synthesis.utils import load_morphology
from axon_synthesis.utils import neurite_to_pts
from axon_synthesis.utils import temp_dir
from axon_synthesis.validation.utils import segment_intersection_lengths

LOGGER = logging.getLogger(__name__)


def _empty_hierarchy() -> OrderedDict:
    """Build 'hierarchy' dict for empty atlas."""
    return OrderedDict(
        [
            ("id", 0),
            ("acronym", "root"),
            ("name", "root"),
            (
                "children",
                [],
            ),
        ]
    )


def empty_atlas(shape, voxel_dimensions, offset, layer_thicknesses=None, tmp_dir=None):
    """Create an empty Atlas."""
    tmp_dir = (
        temp_dir(delete=False) if tmp_dir is None else CleanableDirectory(tmp_dir, exist_ok=True)
    )
    brain_regions = VoxelData(np.zeros(shape, dtype=np.int32), voxel_dimensions, offset)

    if layer_thicknesses is None:
        y_extent = brain_regions.bbox[1, 1] - brain_regions.bbox[0, 1]
        thickness = y_extent / 6
        layer_thicknesses = {str(i): i * thickness for i in range(1, 7)}
    _dump_atlases(brain_regions, layer_thicknesses, tmp_dir.name)

    hierarchy = _empty_hierarchy()
    with (Path(tmp_dir.name) / "hierarchy.json").open(mode="w", encoding="utf-8") as f:
        json.dump(_normalize_hierarchy(hierarchy), f, indent=4)

    return tmp_dir


def update_source_target_regions(brain_regions, hierarchy, source_pt, target_pts, suffix=""):
    """Create brain regions for the given source and targets."""
    hierarchy_df = hierarchy.as_dataframe()

    root = hierarchy_df.loc[hierarchy_df["parent_id"] == -1]
    root_idx = root.index[0]

    new_ind = max(1, hierarchy_df.index.max() + 1)
    hierarchy_df.loc[new_ind, ["acronym", "name", "parent_id", "children_count"]] = [
        "s",
        f"source_region{suffix}",
        root_idx,
        0,
    ]
    brain_regions.raw[brain_regions.positions_to_indices(source_pt)] = new_ind
    new_ind += 1

    for num, i in enumerate(target_pts):
        hierarchy_df.loc[new_ind, ["acronym", "name", "parent_id", "children_count"]] = [
            f"t_{num}",
            f"target_region{suffix}_{num}",
            root_idx,
            0,
        ]
        brain_regions.raw[brain_regions.positions_to_indices(i)] = new_ind
        new_ind += 1

    return brain_regions, RegionMap.from_dataframe(hierarchy_df)


def dummy_preferred_regions(brain_regions, hierarchy, morph, axon_id=0):
    """Create preferred regions as brain regions for the given morphology."""
    hierarchy_df = hierarchy.as_dataframe()

    root = hierarchy_df.loc[hierarchy_df["parent_id"] == -1]

    new_ind = max(1, hierarchy_df.index.max() + 1)
    hierarchy_df.loc[new_ind, ["acronym", "name", "parent_id", "children_count"]] = [
        "dft",
        "dummy_preferred_regions",
        root.index[0],
        0,
    ]

    morph = Morphology(morph)
    axon = get_axons(morph)[axon_id]
    _nodes, edges = neurite_to_pts(axon, keep_section_segments=True, edges_with_coords=True)
    heat_map = segment_intersection_lengths(
        edges, brain_regions.bbox, brain_regions.voxel_dimensions
    )
    brain_regions.raw[np.nonzero(heat_map.raw != 0)] = new_ind

    return brain_regions, RegionMap.from_dataframe(hierarchy_df)


def morph_atlas(
    morph,
    voxel_dimensions=None,
    atlas=None,
    preferred_regions_axon_id=None,
    target_neurite_types=None,
    *,
    export=False,
    tmp_dir=None,
):
    """Create an empty Atlas based on the dimensions of the given morphology."""
    morph = load_morphology(morph, recenter=True)
    if atlas is None:
        bbox = compute_bbox(morph.points[:, COLS.XYZ], absolute_buffer=1)
        extent = bbox[1] - bbox[0]
        shape = (np.clip(extent // voxel_dimensions, 1, np.inf) + 3).astype(int)
        offset = bbox[0]
        atlas_dir = empty_atlas(shape, voxel_dimensions, offset, tmp_dir=tmp_dir)
        atlas_path = Path(atlas_dir.name)
        atlas_config = AtlasConfig(
            atlas_path,
            "brain_regions",
            layer_names=list(range(1, 7)),
            load_region_map=True,
            use_boundary=False,
        )
        atlas = AtlasHelper(atlas_config)
    else:
        atlas_dir = (
            temp_dir(delete=False)
            if tmp_dir is None
            else CleanableDirectory(tmp_dir, exist_ok=True)
        )
        atlas_path = Path(atlas_dir.name)
        atlas_config = evolve(
            atlas.config, path=atlas_path, load_region_map=True, use_boundary=False
        )
        atlas = atlas.copy()
        atlas.config = atlas_config
    brain_regions = atlas.brain_regions
    hierarchy = atlas.region_map

    if target_neurite_types is not None:
        if np.isscalar(target_neurite_types):
            target_neurite_types = [target_neurite_types]

        for num, neurite in enumerate(morph.neurites):
            if neurite.type not in target_neurite_types:
                continue
            source = neurite.root_node.points[0, COLS.XYZ]
            targets = np.array([i.points[0, COLS.XYZ] for i in neurite.iter_sections()])
            brain_regions, hierarchy = update_source_target_regions(
                brain_regions, hierarchy, source, targets, suffix=f"_{neurite.type.name}_{num}"
            )

    if preferred_regions_axon_id is not None:
        brain_regions, hierarchy = dummy_preferred_regions(
            brain_regions, hierarchy, morph, preferred_regions_axon_id
        )

    atlas.brain_regions = brain_regions
    atlas.region_map = hierarchy

    if export:
        atlas.save(atlas_path)
        LOGGER.debug("Exported atlas to '%s'", atlas_path)

    return atlas_dir, atlas



"""Entries of the Command Line Interface dedicated to the validation."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
# LICENSE HEADER MANAGED BY add-license-header

import click

from axon_synthesis.cli.common import parallel_kwargs_to_config
from axon_synthesis.cli.common import parallel_options
from axon_synthesis.cli.input_creation import clustering_parameters_option
from axon_synthesis.cli.synthesis import create_graph_kwargs_to_config
from axon_synthesis.cli.synthesis import create_graph_options
from axon_synthesis.cli.synthesis import output_options
from axon_synthesis.cli.synthesis import outputs_kwargs_to_config
from axon_synthesis.cli.synthesis import post_process_kwargs_to_config
from axon_synthesis.cli.synthesis import post_process_options
from axon_synthesis.cli.utils import GlobalConfig
from axon_synthesis.cli.utils import ListParam
from axon_synthesis.validation.mimic import mimic_axons


@click.command(
    short_help="Synthesize axons mimicking the input ones.",
)
@click.option(
    "--morphology-dir",
    type=click.Path(exists=True, file_okay=False),
    required=True,
    help="The directory containing the input morphologies",
)
@click.option(
    "--workflows",
    type=ListParam(
        schema={
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "basic",
                    "preferred_regions",
                ],
            },
        },
    ),
    help="The mimic workflows to synthesize",
)
@click.option(
    "--voxel-dimensions",
    type=ListParam(
        schema={
            "type": "array",
            "items": {
                "type": "integer",
                "minimum": 1,
            },
        }
    ),
    help="The voxel dimensions used to build the dummy atlas in the 'preferred_regions' workflow",
)
@click.option(
    "--keep-tmp-atlas/--no-keep-tmp-atlas",
    default=False,
    help=("If set to True, the temporary atlas is not removed at the end of the process"),
)
@click.option(
    "--merge-results/--no-merge-results",
    default=True,
    help=("If set to True, the temporary atlas is not removed at the end of the process"),
)
@click.option(
    "--tuft-parameters-file",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to the file containing the tuft parameters given to NeuroTS.",
)
@output_options
@clustering_parameters_option
@create_graph_options
@post_process_options
@parallel_options
@click.pass_obj
def mimic(global_config: GlobalConfig, *_args, **kwargs):
    """The command to synthesize mimicking axons."""
    global_config.to_config(kwargs)
    create_graph_kwargs_to_config(kwargs)
    post_process_kwargs_to_config(kwargs)
    outputs_kwargs_to_config(kwargs)
    parallel_kwargs_to_config(kwargs)
    mimic_axons(**kwargs)



"""Create the edges between the terminals and the obstacles (if any).

This is needed to easily compute a Steiner Tree (Euclidean Steiner Tree is complicated).
"""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

import numpy as np
import pandas as pd
from attrs import define
from attrs import field
from attrs import validators
from scipy.spatial import KDTree
from voxcell import VoxelData

from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import FROM_COORDS_COLS
from axon_synthesis.constants import TARGET_COORDS_COLS
from axon_synthesis.constants import TO_COORDS_COLS
from axon_synthesis.constants import NodeProvider
from axon_synthesis.synthesis.main_trunk.create_graph.plot import plot_triangulation
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_bounding_box_pts
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_depth_penalty
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_intermediate_points
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_orientation_penalty
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_preferred_reward
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_random_points
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_terminal_penalty
from axon_synthesis.synthesis.main_trunk.create_graph.utils import add_voronoi_points
from axon_synthesis.synthesis.main_trunk.create_graph.utils import create_edges
from axon_synthesis.synthesis.main_trunk.create_graph.utils import drop_close_points
from axon_synthesis.synthesis.main_trunk.create_graph.utils import drop_outside_points
from axon_synthesis.typing import FileType
from axon_synthesis.typing import RegionIdsType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import compute_bbox
from axon_synthesis.utils import sublogger


@define
class CreateGraphConfig:
    """Class to store the parameters needed for graph creation.

    Attributes:
        intermediate_number: The number of intermediate points added before Voronoï process.
        min_intermediate_distance: The min distance between two successive intermediate points.
        min_random_point_distance: The min distance used to add random points.
        random_point_bbox_buffer: The distance used to add a buffer around the bbox of the points.
        voronoi_steps: The number of Voronoi steps.
        duplicate_precision: The precision used to detect duplicated points.
        use_orientation_penalty: If set to True, a penalty is added to edges whose direction is not
            radial.
        orientation_penalty_exponent: The exponent used for the orientation penalty.
        orientation_penalty_amplitude: The amplitude of the orientation penalty.
        use_depth_penalty: If set to True, a penalty is added to edges whose direction is not
            parallel to the iso-depth curves.
        depth_penalty_sigma: The sigma used for depth penalty.
        depth_penalty_amplitude: The amplitude of the depth penalty.
        preferred_regions: The list of brain regions in which edge weights are divided by the
            preferring factor.
        preferred_region_min_random_point_distance: The min distance used to pick random points in
            preferred regions.
        preferring_sigma: The sigma used to compute the preferring factor for the given regions.
        preferring_amplitude: The amplitude used to compute the preferring factor for the given
            regions.
        preferred_region_tree: The KDTree object containing the preferred region points.
        use_terminal_penalty: If set to True, a penalty is added to edges that are connected to a
            terminal.
    """

    # Intermediate points
    intermediate_number: int = field(default=5, validator=validators.ge(0))
    min_intermediate_distance: float = field(default=1000, validator=validators.gt(0))

    # Random points
    min_random_point_distance: float | None = field(
        default=None, validator=validators.optional(validators.ge(0))
    )
    random_max_tries: int = field(default=10)
    random_point_bbox_buffer: float = field(default=0, validator=validators.ge(0))

    # Voronoï points
    voronoi_steps: int = field(default=1, validator=validators.ge(0))

    # Duplicated points
    duplicate_precision: float = field(default=1e-3, validator=validators.gt(0))

    # Orientation penalty
    use_orientation_penalty: bool = field(default=True)
    orientation_penalty_exponent: float = field(default=0.1, validator=validators.ge(0))
    orientation_penalty_amplitude: float = field(default=1, validator=validators.gt(0))

    # Depth penalty
    use_depth_penalty: bool = field(default=True)
    depth_penalty_sigma: float = field(default=0.25, validator=validators.gt(0))
    depth_penalty_amplitude: float = field(default=10, validator=validators.gt(0))

    # preferred regions
    preferred_regions: RegionIdsType | None = field(
        default=None,
    )
    preferred_region_min_random_point_distance: float | None = field(
        default=None, validator=validators.optional(validators.ge(0))
    )
    preferring_sigma: float = field(default=100, validator=validators.gt(0))
    preferring_amplitude: float = field(default=1, validator=validators.gt(0))
    preferred_region_tree: KDTree | None = field(default=None)

    # Terminal penalty
    use_terminal_penalty: bool = field(default=False)

    def __attrs_post_init__(self):
        """Check consistency of new instances."""
        if self.preferred_regions is not None and self.preferred_region_min_distance is None:
            msg = (
                "A preferred region list was given without any related distance, please provide "
                "at least one of the following parameters: "
                "'preferred_region_min_random_point_distance', 'min_random_point_distance'",
            )
            raise ValueError(msg)

    @property
    def preferred_region_min_distance(self):
        """Get the min distance used to generate random points in the preferred regions."""
        return (
            self.preferred_region_min_random_point_distance
            if self.preferred_region_min_random_point_distance is not None
            else self.min_random_point_distance
        )

    def pick_preferred_region_random_points(self, rng=None, max_tries: int = 10):
        """Create random points in preferred regions with min distance between them."""
        if self.preferred_region_tree is None:
            msg = (
                "The 'preferred_region_tree' must be computed before picking random points inside "
                "the preferred region"
            )
            raise RuntimeError(msg)
        rng = np.random.default_rng(rng)
        n_fails = 0
        new_pts: list[np.ndarray] = []
        min_dist = self.preferred_region_min_distance
        while n_fails < max_tries:
            xyz = rng.choice(self.preferred_region_tree.data, replace=False, shuffle=False)
            if (
                len(new_pts) == 0
                or np.linalg.norm(
                    xyz - new_pts,
                    axis=1,
                ).min()
                > min_dist
            ):
                new_pts.append(xyz)
                n_fails = 0
            else:
                n_fails += 1

        if len(new_pts) == 0:
            # Ensure at least one point is added (the closest point from the barycenter of the
            # given points)
            new_pts.append(
                self.preferred_region_tree.data[
                    np.argmin(
                        np.linalg.norm(
                            self.preferred_region_tree.data
                            - self.preferred_region_tree.data.mean(axis=0),
                            axis=1,
                        )
                    )
                ]
            )
        return np.hstack([new_pts, np.ones((len(new_pts), 1)) * NodeProvider.attractors])

    def compute_region_tree(self, atlas: AtlasHelper, *, force: bool = False):
        """Compute the preferred region tree using the given Atlas."""
        if self.preferred_regions and (self.preferred_region_tree is None or force):
            preferred_region_points, _missing_ids = atlas.get_region_points(self.preferred_regions)
            self.preferred_region_tree = KDTree(preferred_region_points)

    def load_region_tree_from_file(self, file):
        """Get the preferred region tree from a file."""
        data = np.load(file, allow_pickle=False)
        self.preferred_region_tree = KDTree(data)


def _add_points(source_coords, pts, config, depths, forbidden_regions, rng, logger) -> tuple:
    """Add all points to the source and target points."""
    # Add intermediate points
    all_pts = add_intermediate_points(
        pts,
        source_coords,
        config.min_intermediate_distance,
        config.intermediate_number,
    )

    # Add random points from other regions
    all_pts = add_random_points(
        all_pts,
        config.min_random_point_distance,
        config.random_point_bbox_buffer,
        rng,
        max_tries=config.random_max_tries,
        logger=logger,
    )

    # Add random points from the preferred regions
    if config.preferred_region_tree is not None:
        preferred_region_pts = config.pick_preferred_region_random_points(
            rng=rng, max_tries=config.random_max_tries
        )
        logger.debug("Random points added in the preferred regions: %s", len(preferred_region_pts))
        all_pts = np.concatenate(
            [
                all_pts,
                preferred_region_pts,
            ]
        )

    # Add the bounding box points to ensure a minimum number of points
    all_pts = add_bounding_box_pts(all_pts)

    # Get bbox of points before adding Voronoi points that can go very far
    current_bbox = compute_bbox(all_pts[:, :3], 0.1)
    if depths is not None:
        # Clip the bbox the keep it inside the atlas
        current_bbox[0] = np.clip(current_bbox[0], a_min=depths.bbox[0], a_max=np.inf)
        current_bbox[1] = np.clip(current_bbox[1], a_min=-np.inf, a_max=depths.bbox[1])

    # Add Voronoï points
    all_pts = add_voronoi_points(
        all_pts, config.voronoi_steps, initial_bbox=current_bbox, logger=logger
    )

    if forbidden_regions is not None:
        forbidden_points = forbidden_regions.lookup(all_pts[:, :3])
        all_pts = all_pts[~forbidden_points]

    return all_pts, current_bbox


def one_graph(
    source_coords: np.ndarray,
    target_points: pd.DataFrame,
    config: CreateGraphConfig,
    depths: VoxelData | None = None,
    *,
    forbidden_regions: VoxelData | None = None,
    output_path: FileType | None = None,
    figure_path: FileType | None = None,
    rng: SeedType = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Create the nodes and edges for one axon based on the target points and the atlas."""
    logger = sublogger(logger, __name__)

    rng = np.random.default_rng(rng)

    unique_target_points = target_points.drop_duplicates(subset=["axon_id", "terminal_id"])

    logger.debug("%s points", len(unique_target_points))

    # Terminal points
    pts = np.concatenate([[source_coords], unique_target_points[TARGET_COORDS_COLS].to_numpy()])
    pts = np.hstack(
        (
            pts,
            np.atleast_2d(
                [NodeProvider.source] + [NodeProvider.target] * len(unique_target_points)
            ).T,
        )
    )

    # Add other points
    all_pts, current_bbox = _add_points(
        source_coords, pts, config, depths, forbidden_regions, rng, logger
    )

    # Gather points
    nodes_df = pd.DataFrame(all_pts, columns=[*COORDS_COLS, "NodeProvider"])

    # Mark the source and target points as terminals and the others as intermediates
    nodes_df["is_terminal"] = [True] * len(pts) + [False] * (len(all_pts) - len(pts))

    # Associate the terminal IDs to the nodes
    nodes_df["terminal_id"] = (
        [-1]
        + unique_target_points["terminal_id"].to_numpy().tolist()
        + [-1] * (len(all_pts) - len(pts))
    )

    # Remove close points
    nodes_df = drop_close_points(nodes_df, config.duplicate_precision)

    # Remove outside points
    nodes_df = drop_outside_points(
        nodes_df,
        bbox=current_bbox,
    )

    # Add some intermediate points to ensure there are enough points to build a graph
    if len(nodes_df) <= 5:
        all_pts = add_intermediate_points(
            nodes_df[[*COORDS_COLS, "NodeProvider"]].to_numpy(),
            source_coords,
            0,
            5 - len(nodes_df),
        )
        filling_pts = pd.DataFrame(all_pts[len(nodes_df) :], columns=[*COORDS_COLS, "NodeProvider"])
        filling_pts[["is_terminal", "terminal_id"]] = (False, -1)
        nodes_df = pd.concat([nodes_df, filling_pts], ignore_index=True)

    # Reset index and set IDs
    nodes_df = nodes_df.reset_index(drop=True)
    nodes_df.loc[:, "id"] = nodes_df.index.to_numpy()

    # Create edges using the Delaunay triangulation of the union of the terminals,
    # intermediate and Voronoï points
    edges_df, _ = create_edges(
        nodes_df.loc[:, COORDS_COLS],
        FROM_COORDS_COLS,
        TO_COORDS_COLS,
    )

    logger.debug("%s edges created", len(edges_df))

    # Compute cumulative penalties
    penalties = np.ones(len(edges_df))

    # Increase the weight of edges whose angle with radial direction is close to pi/2
    if config.use_orientation_penalty:
        logger.debug("Add orientation penalties to edge weights")
        penalties *= add_orientation_penalty(
            edges_df,
            FROM_COORDS_COLS,
            TO_COORDS_COLS,
            source_coords,
            config.orientation_penalty_exponent,
            config.orientation_penalty_amplitude,
        )

    # Increase the weight of edges which do not follow an iso-depth curve
    if config.use_depth_penalty and depths is not None:
        logger.debug("Add depth penalties to edge weights")
        penalties *= add_depth_penalty(
            edges_df,
            FROM_COORDS_COLS,
            TO_COORDS_COLS,
            depths,
            config.depth_penalty_sigma,
            config.depth_penalty_amplitude,
        )

    # Reduce the lengths of edges that are close to the preferred regions
    if config.preferred_region_tree is not None:
        logger.debug("Add preferred region rewards to edge weights")
        penalties *= add_preferred_reward(
            edges_df,
            FROM_COORDS_COLS,
            TO_COORDS_COLS,
            config.preferred_region_tree,
            config.preferring_sigma,
            config.preferring_amplitude,
        )

    # TODO: increase weights of more impossible edges?

    # Apply cumulative penalties
    edges_df["weight"] = edges_df["length"] * penalties

    # TODO: Remove points and edges from forbidden regions?

    # Add penalty to edges between two terminals (except if a terminal is only
    # connected to other terminals) in order to ensure the terminals are also terminals
    # in the solution
    # NOTE: This behavior is disabled by default because we don't generate the actual
    # terminals of the tufts with Steiner Tree, we just generate long range trunk that
    # passes near the target points.
    if config.use_terminal_penalty:
        logger.debug("Add terminal penalties to edge weights")
        add_terminal_penalty(edges_df, nodes_df)

    logger.debug("All penalties added to edge weights")

    if output_path is not None:
        logger.debug("Export nodes and edges to %s", output_path)
        Path(output_path).unlink(missing_ok=True)
        Path(output_path).parent.mkdir(parents=True, exist_ok=True)
        nodes_df.to_hdf(output_path, key="nodes")
        edges_df.to_hdf(output_path, key="edges", mode="a")

    if figure_path is not None:
        plot_triangulation(
            edges_df,
            source_coords,
            pts[1:, :3],  # The first one is the source
            figure_path,
            logger=logger,
            attractors=nodes_df.loc[
                nodes_df["NodeProvider"] == NodeProvider.attractors, COORDS_COLS
            ].to_numpy()
            if config.preferred_region_tree is not None
            else None,
        )

    # pylint: disable=unsupported-assignment-operation
    nodes_df["NodeProvider"] = nodes_df.loc[:, "NodeProvider"].map(
        {i: i.name for i in NodeProvider}
    )

    return nodes_df, edges_df



"""Add tufts to Steiner solutions."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from copy import deepcopy
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from morph_tool.converter import single_point_sphere_to_circular_contour
from morphio import SomaType
from morphio.mut import Morphology as MorphIoMorphology
from neurom.core import Morphology
from neurots.generate.section import SectionGrowerPath
from neurots.generate.tree import TreeGrower
from neurots.generate.tree import section_growers
from neurots.morphmath.utils import get_random_point
from plotly.subplots import make_subplots
from plotly_helper.neuron_viewer import NeuronBuilder

from axon_synthesis.synthesis.tuft_properties import TUFT_COORDS_COLS
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import add_camera_sync
from axon_synthesis.utils import build_layout_properties
from axon_synthesis.utils import disable_loggers
from axon_synthesis.utils import sublogger


class SectionGrowerBoundary(SectionGrowerPath):
    """Section grower that will not cross the given boundary."""

    def next_point(self, current_point) -> tuple:
        """Compute the next point and update it if close to the boundary."""
        boundary = self.context.get("boundary", None) if self.context is not None else None
        if boundary is None:
            return super().next_point(current_point)

        direction = (
            self.params.targeting * self.direction
            + self.params.randomness * get_random_point(random_generator=self._rng)
            + self.params.history * self.history()
        )

        direction = direction / np.linalg.norm(direction)  # From NeuroTS
        seg_length = self.step_size_distribution.draw_positive()  # From NeuroTS

        # Check where is the next point compared to the boundaries
        voxel_pos, voxel_idx = np.modf(
            (current_point - boundary.offset) / boundary.voxel_dimensions
        )
        boundary_vec = boundary.raw[tuple(voxel_idx.astype(int))]
        distance = np.linalg.norm(boundary_vec)

        # Refine boundary vector with intra-pixel position (simple linear interpolation)
        actual_boundary_vec = boundary_vec * (
            1 + np.dot(voxel_pos, boundary_vec) / max(1e-3, distance**2)
        )
        actual_distance = np.linalg.norm(actual_boundary_vec)

        # Compute the boundary attenuation component
        if actual_distance <= self.context["boundary_max_distance"]:
            boundary_direction = actual_boundary_vec / actual_distance
            attenuation = boundary_direction * np.exp(
                -self.context["boundary_scale_coeff"] * actual_distance
            )
            direction += attenuation
            direction = direction / np.linalg.norm(direction)

        # Update the next point
        next_point = current_point + seg_length * direction  # From NeuroTS
        self.update_pathlength(seg_length)  # From NeuroTS

        return next_point, direction


section_growers["path_distances+boundary"] = SectionGrowerBoundary


def plot_tuft(morph, title, output_path, initial_morph=None, morph_title=None, logger=None):
    """Plot the given morphology.

    If `initial_morph` is not None then the given morphology is also plotted for comparison.
    """
    morph = Morphology(morph)
    fig_builder = NeuronBuilder(morph, "3d", line_width=4, title=title)
    fig_data = [fig_builder.get_figure()["data"]]
    left_title = "Morphology with tufts"

    if initial_morph is not None:
        if morph_title is None:
            morph_title = "Raw morphology"

        fig = make_subplots(
            cols=2,
            specs=[[{"type": "scene"}, {"type": "scene"}]],
            subplot_titles=[left_title, morph_title],
        )

        if initial_morph.root_sections:
            fig.add_traces(
                NeuronBuilder(initial_morph, "3d", line_width=4, title=title).get_figure()["data"]
            )
        else:
            fig_builder = fig.add_traces(
                go.Scatter3d(
                    x=[initial_morph.soma.center[0]],
                    y=[initial_morph.soma.center[1]],
                    z=[initial_morph.soma.center[2]],
                    marker={"color": "black", "size": 4},
                    mode="markers",
                    name="Soma",
                )
            )
    else:
        fig = make_subplots(cols=1, specs=[[{"type": "scene"}]], subplot_titles=[left_title])

    for col_num, data in enumerate(fig_data):
        fig.add_traces(data, rows=[1] * len(data), cols=[col_num + 1] * len(data))

    layout_props = build_layout_properties(morph.points, 0.5)

    fig.update_scenes(layout_props)
    fig.update_layout(title=morph.name)

    # Export figure
    Path(output_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(output_path)

    if initial_morph is not None:
        add_camera_sync(output_path)

    if logger is not None:
        logger.info("Exported figure to %s", output_path)


def build_and_graft_tufts(
    morph: Morphology,
    tuft_properties: pd.DataFrame,
    parameters: dict,
    distributions: dict,
    *,
    context=None,
    output_dir: FileType | None = None,
    figure_dir: FileType | None = None,
    initial_morph: Morphology | None = None,
    rng: SeedType = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Build the tufts and graft them to the given morphology.

    .. warning::
        The directories passed to ``output_dir`` and ``figure_dir`` should already exist.
    """
    logger = sublogger(logger, __name__)

    if output_dir is not None:
        output_dir = Path(output_dir)
    if figure_dir is not None:
        figure_dir = Path(figure_dir)

    rng = np.random.default_rng(rng)

    for _, row in tuft_properties.iterrows():
        # Create specific parameters
        params = deepcopy(parameters)
        tuft_orientation = np.dot(row["target_orientation"], row["tuft_orientation"])
        params["axon"]["orientation"]["values"]["orientations"] = [tuft_orientation]
        logger.debug("Tuft orientation: %s", tuft_orientation)

        # Create specific distributions
        distrib = deepcopy(distributions)
        distrib["axon"]["persistence_diagram"] = [
            row["barcode"],
        ]
        logger.debug("Tuft barcode: %s", row["barcode"])

        initial_point = [row[col] for col in TUFT_COORDS_COLS]
        logger.debug("Tuft start point: %s", initial_point)

        # Grow a tuft
        new_morph = MorphIoMorphology()

        grower = TreeGrower(
            new_morph,
            initial_direction=tuft_orientation,
            initial_point=initial_point,
            parameters=params["axon"],
            distributions=distrib["axon"],
            context=context,
            random_generator=rng,
        )
        while not grower.end():
            grower.next_point()

        filename = f"{row['morphology']}_{row['axon_id']}_{row['terminal_id']}"
        if output_dir is not None:
            new_morph.soma.points = [initial_point]
            new_morph.soma.diameters = [0.5]
            new_morph.soma.type = SomaType.SOMA_SINGLE_POINT
            with disable_loggers("morph_tool.converter"):
                single_point_sphere_to_circular_contour(new_morph)
            output_path = (output_dir / filename).with_suffix(".h5")
            Path(output_path).parent.mkdir(parents=True, exist_ok=True)
            new_morph.write(output_path)

        if figure_dir is not None:
            plot_tuft(
                new_morph,
                filename,
                (figure_dir / filename).with_suffix(".html"),
                initial_morph,
                logger=logger,
            )

        # Graft the tuft to the current terminal
        sec = morph.section(row["section_id"])
        if row["use_parent"]:
            sec = sec.parent
        sec.append_section(new_morph.root_sections[0], recursive=True)



"""Base of the synthesis modules."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import json
import logging
import os
from pathlib import Path
from typing import ClassVar

import dask.distributed
import morphio
import numpy as np
import pandas as pd
from attrs import asdict
from attrs import converters
from attrs import define
from attrs import evolve
from attrs import field
from attrs import validators
from neurom import NeuriteType
from neurom.core import Morphology
from neurom.geom.transform import Translation
from voxcell.cell_collection import CellCollection
from wurlitzer import STDOUT
from wurlitzer import pipes

try:
    import dask_mpi
    from mpi4py import MPI

    mpi_enabled = True
except ImportError:
    mpi_enabled = False

from axon_synthesis import __version__
from axon_synthesis.atlas import AtlasConfig
from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.base_path_builder import FILE_SELECTION
from axon_synthesis.base_path_builder import BasePathBuilder
from axon_synthesis.constants import AXON_GRAFTING_POINT_HDF_GROUP
from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import DEFAULT_OUTPUT_PATH
from axon_synthesis.constants import TARGET_COORDS_COLS
from axon_synthesis.inputs import Inputs
from axon_synthesis.synthesis.add_tufts import build_and_graft_tufts
from axon_synthesis.synthesis.main_trunk.create_graph import CreateGraphConfig
from axon_synthesis.synthesis.main_trunk.create_graph import one_graph
from axon_synthesis.synthesis.main_trunk.post_process import PostProcessConfig
from axon_synthesis.synthesis.main_trunk.post_process import post_process_trunk
from axon_synthesis.synthesis.main_trunk.steiner_morphology import build_and_graft_trunk
from axon_synthesis.synthesis.main_trunk.steiner_tree import compute_solution
from axon_synthesis.synthesis.outputs import OutputConfig
from axon_synthesis.synthesis.outputs import Outputs
from axon_synthesis.synthesis.plot import plot_final_morph
from axon_synthesis.synthesis.plot import plot_target_points
from axon_synthesis.synthesis.source_points import SOURCE_COORDS_COLS
from axon_synthesis.synthesis.source_points import set_source_points
from axon_synthesis.synthesis.target_points import get_target_points
from axon_synthesis.synthesis.tuft_properties import pick_barcodes
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import MorphNameAdapter
from axon_synthesis.utils import ParallelConfig
from axon_synthesis.utils import create_dask_dataframe
from axon_synthesis.utils import disable_distributed_loggers
from axon_synthesis.utils import export_morph_edges
from axon_synthesis.utils import load_morphology
from axon_synthesis.utils import permanently_disable_distributed_loggers
from axon_synthesis.utils import save_morphology
from axon_synthesis.utils import setup_logger

LOGGER = logging.getLogger(__name__)


class SynthesisError(Exception):
    """Exception raised by axon synthesis."""


@define
class SynthesisConfig:
    """Class to store the synthesis configuration.

    Attributes:
        morphology_dir: Path to the directory containing the input morphologies.
        morphology_data_file: Path to the MVD3 or SONATA file containing the morphology data.
        axon_grafting_points_file: Path to the HDF5 file containing the section IDs where the axons
            should be grafted in the input morphologies (axons are grafted to the soma if not
            provided).
        input_dir: Path to the directory containing the inputs.
        population_probabilities_file: Path to the file containing the population probabilities.
        projection_probabilities_file: Path to the file containing the projection probabilities.
        population_tuft_number_file: Path to the file containing the distribution of number of tufts
            per pathway.
        trunk_properties_file: Path to the file containing the trunk properties.
        tuft_properties_file: Path to the file containing the tuft properties.
        tuft_distributions_file: Path to the file containing the distributions used for tuft
            synthesis.
        tuft_parameters_file: Path to the file containing the parameters used for tuft synthesis.
        tuft_boundary_max_distance: Maximum distance used for the calculation of the attenuation
            component near the boundary.
        tuft_boundary_scale_coeff: Coefficient used in the calculation of the attenuation
            component near the boundary.
        rebuild_existing_axons: If set to True, all existing axons are removed before synthesis.
        target_max_tries: The maximum number of tries for picking target points.
    """

    morphology_dir: FileType = field(converter=Path)
    morphology_data_file: FileType = field(converter=Path)
    axon_grafting_points_file: FileType | None = field(
        default=None, converter=converters.optional(Path)
    )
    input_dir: FileType | None = field(default=None, converter=converters.optional(Path))
    population_probabilities_file: FileType | None = field(
        default=None, converter=converters.optional(Path)
    )
    projection_probabilities_file: FileType | None = field(
        default=None, converter=converters.optional(Path)
    )
    population_tuft_number_file: FileType | None = field(
        default=None, converter=converters.optional(Path)
    )
    trunk_properties_file: FileType | None = field(
        default=None, converter=converters.optional(Path)
    )
    tuft_properties_file: FileType | None = field(default=None, converter=converters.optional(Path))
    tuft_distributions_file: FileType | None = field(
        default=None, converter=converters.optional(Path)
    )
    tuft_parameters_file: FileType | None = field(default=None, converter=converters.optional(Path))
    tuft_boundary_max_distance: float = field(default=200)
    tuft_boundary_scale_coeff: float | None = field(default=None)
    rebuild_existing_axons: bool = field(default=False)

    target_max_tries: int = field(default=10, validator=validators.gt(0))


def load_axon_grafting_points(
    path: FileType | None = None, key: str = AXON_GRAFTING_POINT_HDF_GROUP
):
    """Load the axon mapping from the given file."""
    cols = ["morphology", "grafting_section_id"]
    if path is not None:
        path = Path(path)
        if path.exists():
            mapping = pd.read_hdf(path, key)
            if set(cols).difference(mapping.columns):
                msg = f"The DataFrame loaded from '{path}' must contain the {cols} columns."
                raise ValueError(msg)
            return mapping.fillna({"grafting_section_id": -1}).astype(  # type: ignore[call-overload]
                {"grafting_section_id": int}
            )
    return None


def remove_existing_axons(morph):
    """Remove all existing axons from a given morphology."""
    for i in morph.root_sections:
        if i.type == NeuriteType.axon:
            morph.delete_section(i)


def one_axon_paths(outputs, morph_file_name, figure_file_name):
    """Create a BasePathBuilder object to store the paths needed for a specific axon."""

    class AxonPaths(BasePathBuilder):
        """Class to store the synthesis outputs of one specific axon."""

        _filenames: ClassVar[dict] = {
            "FIGURE_FILE_NAME": figure_file_name,
            "GRAPH_CREATION_FIGURE": (outputs.GRAPH_CREATION_FIGURES / figure_file_name)
            if outputs.GRAPH_CREATION_FIGURES is not None
            else None,
            "GRAPH_CREATION_DATA": (outputs.GRAPH_CREATION_DATA / morph_file_name)
            if outputs.GRAPH_CREATION_DATA is not None
            else None,
            "MAIN_TRUNK_FIGURE": (outputs.MAIN_TRUNK_FIGURES / figure_file_name)
            if outputs.MAIN_TRUNK_FIGURES is not None
            else None,
            "MAIN_TRUNK_MORPHOLOGY": (outputs.MAIN_TRUNK_MORPHOLOGIES / morph_file_name)
            if outputs.MAIN_TRUNK_MORPHOLOGIES is not None
            else None,
            "MORPH_FILE_NAME": morph_file_name,
            "POSTPROCESS_TRUNK_FIGURE": (outputs.POSTPROCESS_TRUNK_FIGURES / figure_file_name)
            if outputs.POSTPROCESS_TRUNK_FIGURES is not None
            else None,
            "POSTPROCESS_TRUNK_MORPHOLOGY": (
                outputs.POSTPROCESS_TRUNK_MORPHOLOGIES / morph_file_name
            )
            if outputs.POSTPROCESS_TRUNK_MORPHOLOGIES is not None
            else None,
            "STEINER_TREE_SOLUTION_FIGURE": (
                outputs.STEINER_TREE_SOLUTION_FIGURES / figure_file_name
            )
            if outputs.STEINER_TREE_SOLUTION_FIGURES is not None
            else None,
            "STEINER_TREE_SOLUTION_NODES": (
                outputs.STEINER_TREE_SOLUTIONS / (Path(morph_file_name).stem + "_nodes.feather")
            )
            if outputs.STEINER_TREE_SOLUTIONS is not None
            else None,
            "STEINER_TREE_SOLUTION_EDGES": (
                outputs.STEINER_TREE_SOLUTIONS / (Path(morph_file_name).stem + "_edges.feather")
            )
            if outputs.STEINER_TREE_SOLUTIONS is not None
            else None,
            "TARGET_POINT_FIGURE": (outputs.TARGET_POINT_FIGURES / figure_file_name)
            if outputs.TARGET_POINT_FIGURES is not None
            else None,
            "TUFT_FIGURES": outputs.TUFT_FIGURES if outputs.TUFT_FIGURES is not None else None,
            "TUFT_MORPHOLOGIES": outputs.TUFT_MORPHOLOGIES
            if outputs.TUFT_MORPHOLOGIES is not None
            else None,
        }

        _optional_keys: ClassVar[set[str]] = {
            "GRAPH_CREATION_FIGURE",
            "GRAPH_CREATION_DATA",
            "MAIN_TRUNK_FIGURE",
            "MAIN_TRUNK_MORPHOLOGY",
            "POSTPROCESS_TRUNK_FIGURE",
            "POSTPROCESS_TRUNK_MORPHOLOGY",
            "STEINER_TREE_SOLUTION_FIGURE",
            "STEINER_TREE_SOLUTION_NODES",
            "STEINER_TREE_SOLUTION_EDGES",
            "TARGET_POINT_FIGURE",
            "TUFT_FIGURES",
            "TUFT_MORPHOLOGIES",
        }

    return AxonPaths("")


@disable_distributed_loggers()
def _init_parallel(
    parallel_config: ParallelConfig, *, mpi_only: bool = False, max_nb_processes: int | None = None
) -> dask.distributed.Client | None:
    """Initialize MPI workers if required or get the number of available processes."""
    if mpi_only and not parallel_config.use_mpi:
        return None

    if parallel_config.nb_processes == 0 and not parallel_config.use_mpi:
        return None

    # Define a default configuration to disable some dask.distributed things
    default_dask_config = {
        "distributed": {
            "worker": {
                "use_file_locking": False,
                "memory": {
                    "target": False,
                    "spill": False,
                    "pause": 0.8,
                    "terminate": 0.95,
                },
                "profile": {
                    "enabled": False,
                    "interval": "10s",
                    "cycle": "10m",
                },
            },
            "admin": {
                "tick": {
                    "limit": "1h",
                },
            },
            "comm": {
                "retry": {
                    "count": 10,
                },
                "timeouts": {
                    "connect": 30,
                },
            },
        },
        "dataframe": {
            "convert_string": False,
        },
    }

    # Merge the default config with the existing config (keep conflicting values from defaults)
    new_dask_config = dask.config.merge(dask.config.config, default_dask_config)

    # Get temporary-directory from environment variables
    _tmp = os.environ.get("SHMDIR", None) or os.environ.get("TMPDIR", None)
    if _tmp is not None:
        new_dask_config["temporary-directory"] = _tmp

    # Merge the config with the one given as argument
    if parallel_config.dask_config is not None:
        new_dask_config = dask.config.merge(new_dask_config, parallel_config.dask_config)

    # Set the dask config
    dask.config.set(new_dask_config)

    if parallel_config.use_mpi:  # pragma: no cover
        if not mpi_enabled:
            msg = (
                "The 'dask' and 'mpi4py' packages must be installed when using the MPI parallel "
                "backend"
            )
            raise ImportError(msg)
        dask_mpi.initialize(dashboard=False)
        comm = MPI.COMM_WORLD  # pylint: disable=c-extension-no-member
        parallel_config.nb_processes = comm.Get_size()
        client_kwargs = {}
        LOGGER.debug(
            "Initializing parallel workers using MPI (%s workers found)",
            parallel_config.nb_processes,
        )
    elif mpi_only:
        return None
    else:
        cpu_count: int = os.cpu_count() or 0
        parallel_config.nb_processes = min(
            parallel_config.nb_processes if parallel_config.nb_processes is not None else cpu_count,
            max_nb_processes if max_nb_processes is not None else cpu_count,
        )

        client_kwargs = {
            "n_workers": parallel_config.nb_processes,
            "dashboard_address": None,
        }
        LOGGER.debug("Initializing parallel workers using the following config: %s", client_kwargs)

    LOGGER.debug("Using the following dask configuration: %s", json.dumps(dask.config.config))

    # This is needed to make dask aware of the workers
    client = dask.distributed.Client(**client_kwargs)

    # Setup logging in workers
    client.run(setup_logger, level=logging.getLevelName(LOGGER.getEffectiveLevel()))
    if LOGGER.getEffectiveLevel() > logging.DEBUG:
        client.run(permanently_disable_distributed_loggers)

    return client


def check_targets(terminals, morph_name):
    """Check that the given terminals have at least one target."""
    missing_target = terminals.groupby("axon_id")["target_population_id"].apply(
        lambda df: df.isna().all()
    )
    if missing_target.any():
        axon_ids = [str(i) for i in missing_target.loc[missing_target].index.tolist()]
        msg = (
            f"Can not synthesize the morphology {morph_name} because no target "
            f"point could be found for the axon(s): {', '.join(axon_ids)}"
        )
        raise SynthesisError(msg)


def synthesize_one_morph_axons(
    morph_terminals,
    inputs,
    outputs,
    create_graph_config,
    post_process_config,
    *,
    tuft_context=None,
    rebuild_existing_axons=False,
    logger=None,
):
    """Synthesize the axons of one morphology."""
    if logger is None:
        logger = LOGGER
    morph_name = morph_terminals.name
    morph_custom_logger = MorphNameAdapter(logger, extra={"morph_name": morph_name})
    try:
        check_targets(morph_terminals, morph_name)

        morph = load_morphology(morph_terminals["morph_file"].iloc[0])

        # Translate the morphology to its position in the atlas
        morph = morph.transform(
            Translation(morph_terminals[COORDS_COLS].iloc[0].to_numpy() - morph.soma.center)
        )

        morph.name = morph_name

        initial_morph = (
            Morphology(morph, name=morph.name)
            if outputs.FINAL_FIGURES is not None or outputs.MORPHOLOGIES_EDGES is not None
            else None
        )

        if rebuild_existing_axons:
            # Remove existing axons
            morph_custom_logger.info("Removing existing axons")
            remove_existing_axons(morph)

        for axon_id, axon_terminals in morph_terminals.groupby("axon_id"):
            # Create a custom logger to add the morph name and axon ID in the log entries
            axon_custom_logger = MorphNameAdapter(
                logger, extra={"morph_name": morph_name, "axon_id": axon_id}
            )

            seed = axon_terminals["seed"].to_numpy()[0]
            morph_custom_logger.info("Starting synthesis of axon %s (seed=%s)", axon_id, seed)
            rng = np.random.default_rng(seed)

            axon_paths = one_axon_paths(
                outputs,
                f"{morph_name}_{axon_id}.h5",
                f"{morph_name}_{axon_id}.html",
            )

            # Create a plot for the initial morph with source and target points
            if axon_paths.TARGET_POINT_FIGURE is not None:
                plot_target_points(
                    initial_morph,
                    axon_terminals[SOURCE_COORDS_COLS].to_numpy()[0],
                    axon_terminals[TARGET_COORDS_COLS].to_numpy(),
                    axon_paths.TARGET_POINT_FIGURE,
                    logger=axon_custom_logger,
                )

            # Create the graph for the current axon
            nodes, edges = one_graph(
                axon_terminals[SOURCE_COORDS_COLS].to_numpy()[0],
                axon_terminals,
                create_graph_config,
                depths=inputs.atlas.depths if inputs.atlas is not None else None,
                forbidden_regions=inputs.atlas.forbidden_regions
                if inputs.atlas is not None
                else None,
                rng=rng,
                output_path=axon_paths.GRAPH_CREATION_DATA,
                figure_path=axon_paths.GRAPH_CREATION_FIGURE,
                logger=axon_custom_logger,
            )

            # Build the Steiner Tree for the current axon
            _, solution_edges = compute_solution(
                nodes,
                edges,
                output_path_nodes=axon_paths.STEINER_TREE_SOLUTION_NODES,
                output_path_edges=axon_paths.STEINER_TREE_SOLUTION_EDGES,
                figure_path=axon_paths.STEINER_TREE_SOLUTION_FIGURE,
                logger=axon_custom_logger,
            )

            # Create the trunk morphology
            trunk_section_id = build_and_graft_trunk(
                morph,
                axon_terminals["grafting_section_id"].to_numpy()[0],
                solution_edges,
                output_path=axon_paths.MAIN_TRUNK_MORPHOLOGY,
                figure_path=axon_paths.MAIN_TRUNK_FIGURE,
                initial_morph=initial_morph,
                logger=axon_custom_logger,
            )

            # Choose a barcode for each tuft of the current axon
            barcodes = pick_barcodes(
                axon_terminals,
                solution_edges,
                inputs.clustering_data.tuft_properties,
                rng=rng,
                logger=axon_custom_logger,
            )

            # Post-process the trunk
            post_process_trunk(
                morph,
                trunk_section_id,
                inputs.clustering_data.trunk_properties,
                barcodes,
                post_process_config,
                rng=rng,
                output_path=axon_paths.POSTPROCESS_TRUNK_MORPHOLOGY,
                figure_path=axon_paths.POSTPROCESS_TRUNK_FIGURE,
                logger=axon_custom_logger,
            )

            # Create the tufts for each barcode
            build_and_graft_tufts(
                morph,
                barcodes,
                inputs.tuft_parameters,
                inputs.tuft_distributions,
                context=tuft_context,
                rng=rng,
                output_dir=axon_paths.TUFT_MORPHOLOGIES,
                figure_dir=axon_paths.TUFT_FIGURES,
                initial_morph=initial_morph,
                logger=axon_custom_logger,
            )

            # TODO: Diametrize the synthesized axon

        # Remove unifurcations of the final morphology
        morphio.set_maximum_warnings(0)
        with pipes(axon_custom_logger, stderr=STDOUT):
            morph.remove_unifurcations()

        # Export the final morph
        final_morph_path = outputs.MORPHOLOGIES / f"{morph_name}.h5"
        save_morphology(
            morph,
            final_morph_path,
            msg=f"Export synthesized morphology to {final_morph_path}",
            logger=morph_custom_logger,
        )

        # Create a plot for the final morph
        if outputs.FINAL_FIGURES is not None:
            plot_final_morph(
                morph,
                morph_terminals,
                outputs.FINAL_FIGURES / f"{morph_name}.html",
                initial_morph,
                logger=morph_custom_logger,
            )

        # Export the final morph as edges
        if outputs.MORPHOLOGIES_EDGES is not None:
            export_morph_edges(
                initial_morph,
                outputs.MORPHOLOGIES_EDGES / f"{morph_name}_initial.csv",
                logger=morph_custom_logger,
            )
            export_morph_edges(
                morph,
                outputs.MORPHOLOGIES_EDGES / f"{morph_name}_final.csv",
                logger=morph_custom_logger,
            )
        res = {
            "file_path": final_morph_path,
            "debug_infos": None,
        }
    except Exception as exc:
        morph_custom_logger.exception(
            "Skip the morphology because of the following error:",
        )
        res = {
            "file_path": None,
            "debug_infos": str(exc),
        }
    return pd.Series(res, dtype=object)


def synthesize_group_morph_axons(
    df: pd.DataFrame, inputs: Inputs, synthesis_config: SynthesisConfig, **func_kwargs
) -> pd.DataFrame:
    """Synthesize all axons of each morphology."""
    if "target_orientation" not in df.columns:
        df["target_orientation"] = np.repeat([np.eye(3)], len(df), axis=0).tolist()
        if inputs.atlas is not None:
            # Update the context of the section grower used for the tufts
            func_kwargs["tuft_context"] = {
                "boundary": inputs.atlas.boundary,
                "boundary_max_distance": synthesis_config.tuft_boundary_max_distance,
                "boundary_scale_coeff": synthesis_config.tuft_boundary_scale_coeff
                if synthesis_config.tuft_boundary_scale_coeff is not None
                else synthesis_config.tuft_boundary_max_distance / 10000,
            }

            mask = ~df["target_population_id"].isna()
            try:
                target_orientations = inputs.atlas.orientations.lookup(
                    df.loc[mask, TARGET_COORDS_COLS].to_numpy()
                )
            except Exception as exc:
                msg = (
                    "Could not get orientations of the target points from the atlas (see complete "
                    "backtrace for details)"
                )
                raise RuntimeError(msg) from exc
            missing_orientations = np.isnan(target_orientations).any(axis=(1, 2))
            if missing_orientations.any():
                target_orientations[missing_orientations] = np.repeat(
                    [np.eye(3)], missing_orientations.sum(), axis=0
                )
            df.loc[df.loc[mask].index, "target_orientation"] = target_orientations.tolist()

    return df.groupby("morphology", group_keys=True).apply(
        lambda group: synthesize_one_morph_axons(group, inputs=inputs, **func_kwargs)
    )


def _partition_wrapper(
    df: pd.DataFrame,
    input_path: FileType,
    synthesis_config: SynthesisConfig,
    atlas_config: AtlasConfig | None,
    **func_kwargs,
) -> pd.DataFrame:
    """Wrapper to process dask partitions."""
    inputs = Inputs(input_path)
    if atlas_config is not None:
        atlas_config.load_region_map = False
        inputs.load_atlas(atlas_config)
    synthesis_config_dict = asdict(synthesis_config)
    inputs.load_clustering_data(**synthesis_config_dict)
    inputs.load_probabilities(**synthesis_config_dict)
    inputs.load_tuft_params_and_distrs(**synthesis_config_dict)

    return synthesize_group_morph_axons(
        df.copy(deep=False), inputs=inputs, synthesis_config=synthesis_config, **func_kwargs
    )


def synthesize_axons(
    config: SynthesisConfig,
    output_config: OutputConfig | None = None,
    *,
    atlas_config: AtlasConfig | AtlasHelper | None = None,
    create_graph_config: CreateGraphConfig | None = None,
    post_process_config: PostProcessConfig | None = None,
    rng: SeedType = None,
    parallel_config: ParallelConfig | None = None,
):  # pylint: disable=too-many-arguments
    """Synthesize the long-range axons.

    Args:
        config: The configuration used for synthesis.
        output_config: The config used to adjust the outputs.
        atlas_config: The config used to load the Atlas.
        create_graph_config: The config used to create the graph.
        post_process_config: The config used to post-process the long-range trunk.
        rng: The random seed or the random generator.
        parallel_config: The configuration for parallel computation.
    """
    LOGGER.info("Using the following version of axon-synthesis: %s", __version__)
    parallel_config = ParallelConfig() if parallel_config is None else evolve(parallel_config)
    _parallel_client = _init_parallel(parallel_config)

    rng = np.random.default_rng(rng)
    outputs = Outputs(
        output_config if output_config is not None else OutputConfig(DEFAULT_OUTPUT_PATH),
        create=True,
    )
    outputs.create_dirs(file_selection=FILE_SELECTION.REQUIRED_ONLY)

    # Load all inputs
    if atlas_config is not None and isinstance(atlas_config, AtlasConfig):
        atlas_config.load_region_map = True
    inputs = Inputs.load(config.input_dir, atlas_config=atlas_config, **asdict(config))  # pylint: disable=not-a-mapping

    # Load the cell collection
    cells_df = CellCollection.load(config.morphology_data_file).as_dataframe()

    # Load the axon grafting points
    axon_grafting_points = load_axon_grafting_points(config.axon_grafting_points_file)

    # Ensure the graph creation config is complete
    if create_graph_config is None:
        create_graph_config = CreateGraphConfig()
    if inputs.atlas is not None:
        create_graph_config.compute_region_tree(inputs.atlas)
    LOGGER.debug("The following config is used for graph creation: %s", create_graph_config)

    if post_process_config is None:
        post_process_config = PostProcessConfig()

    # Get source points for all axons
    source_points = set_source_points(
        cells_df,
        inputs.atlas,
        config.morphology_dir,
        inputs.population_probabilities,
        axon_grafting_points,
        rng=rng,
        rebuild_existing_axons=config.rebuild_existing_axons,
        logger=LOGGER,
    )

    # Find targets for all axons
    target_points = get_target_points(
        source_points,
        inputs.projection_probabilities,
        inputs.tuft_number,
        create_graph_config.duplicate_precision,
        atlas=inputs.atlas,
        brain_regions_masks=inputs.brain_regions_mask_file,
        rng=rng,
        max_tries=config.target_max_tries,
        output_path=outputs.TARGET_POINTS,
        logger=LOGGER,
    )
    if config.rebuild_existing_axons:
        # If the existing axons are rebuilt all the new axons will be grafted to the soma
        target_points["grafting_section_id"] = -1

    if "seed" not in target_points.columns:
        # If no random seed is provided in the data, new ones are created for each morphologys
        target_points = target_points.merge(
            pd.Series(1, index=target_points["morphology"].unique(), name="seed")
            .sample(frac=1, random_state=rng)
            .cumsum()
            - 1,
            left_on="morphology",
            right_index=True,
            how="left",
        )

    func_kwargs = {
        "outputs": outputs,
        "create_graph_config": create_graph_config,
        "post_process_config": post_process_config,
        "rebuild_existing_axons": config.rebuild_existing_axons,
        "logger": LOGGER,
    }

    if parallel_config.nb_processes == 0:
        LOGGER.info("Start computation")
        computed = synthesize_group_morph_axons(target_points, inputs, config, **func_kwargs)
    else:
        LOGGER.info("Start parallel computation using %s workers", parallel_config.nb_processes)
        ddf = create_dask_dataframe(
            target_points, parallel_config.nb_processes, group_col="morphology"
        )
        meta = pd.DataFrame(
            {
                name: pd.Series(dtype="object")
                for name in [
                    "file_path",
                    "debug_infos",
                ]
            }
        )
        future = ddf.map_partitions(
            _partition_wrapper,
            meta=meta,
            input_path=inputs.path,
            synthesis_config=config,
            atlas_config=inputs.atlas.config if inputs.atlas is not None else None,
            **func_kwargs,
        )
        computed = future.compute()

    LOGGER.info("Format results")
    res = target_points.merge(computed, left_on="morphology", right_index=True, how="left")
    if res["file_path"].isna().any():
        LOGGER.error(
            "The following morphologies could not be synthesized (see the logs for details): %s",
            sorted(res.loc[res["file_path"].isna(), "morphology"].unique().tolist()),
        )
    LOGGER.info(
        "Synthesized %s morphologies on %s",
        len(res.loc[~res["file_path"].isna(), "morphology"].unique()),
        len(source_points["morphology"].unique()),
    )
    try:
        if _parallel_client is not None:
            with disable_distributed_loggers():
                _parallel_client.close()
                del _parallel_client
    except Exception:
        LOGGER.exception(
            "The internal parallel client may not have been closed properly because of this error"
        )
    return res
